# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_pulseekko is a module developed to ingest Sensors & Software pulseEKKO GPR data
adapted from ImpDAR/lib/load/load_pulse_ekko.py
"""
### imports ###
from ragu.radar import garlic
from ragu.nav import navparse
import os,sys,struct,datetime,re
import numpy as np

class TraceHeaders:
    """Class used internally to handle pulse-ekko headers."""

    def __init__(self, tnum):
        """Create a container for all the trace headers."""
        self.header_index = 0
        self.trace_numbers = np.zeros((1, tnum))
        self.positions = np.zeros((1, tnum))
        self.points_per_trace = np.zeros((1, tnum))
        self.topography = np.zeros((1, tnum))
        self.bytes_per_point = np.zeros((1, tnum))
        self.n_stackes = np.zeros((1, tnum))
        self.time_window = np.zeros((1, tnum))
        self.pos = np.zeros((3, tnum))
        self.receive = np.zeros((3, tnum))
        self.transmit = np.zeros((3, tnum))
        self.tz_adjustment = np.zeros((1, tnum))
        self.zero_flag = np.zeros((1, tnum))
        self.time_of_day = np.zeros((1, tnum))
        self.comment_flag = np.zeros((1, tnum))
        self.comment = ['' for i in range(tnum)]

    def get_header(self, offset, f_lines):
        """Get the header information for a single trace."""
        header = struct.unpack('<25f', f_lines[offset: offset + 25 * 4])
        comment = struct.unpack('<28c',
                                f_lines[offset + 25 * 4: offset + 25 * 4 + 28])
        self.trace_numbers[0, self.header_index] = header[0]
        self.positions[0, self.header_index] = header[1]
        self.points_per_trace[0, self.header_index] = header[2]
        self.topography[0, self.header_index] = header[3]
        self.bytes_per_point[0, self.header_index] = header[5]
        self.n_stackes[0, self.header_index] = header[7]
        self.time_window[0, self.header_index] = header[8]
        self.pos[0, self.header_index] = header[9]
        self.pos[1, self.header_index] = header[11]
        self.pos[2, self.header_index] = header[13]
        self.receive[0, self.header_index] = header[14]
        self.receive[1, self.header_index] = header[15]
        self.receive[2, self.header_index] = header[16]
        self.transmit[0, self.header_index] = header[17]
        self.transmit[1, self.header_index] = header[18]
        self.transmit[2, self.header_index] = header[19]
        self.tz_adjustment[0, self.header_index] = header[20]
        self.zero_flag[0, self.header_index] = header[21]
        self.time_of_day[0, self.header_index] = header[23]
        self.comment_flag[0, self.header_index] = header[24]
        self.comment[self.header_index] = str(comment[0])
        self.header_index += 1


def partition_project_file(fpath, navcrs, body):
    """Separate profiles.
    The new pulse ekko dvl writes 'project' files with all the profiles stored
    together. We want to break them out into all the .HD header files and .DT1
    data files.
    Parameters
    ----------
    fpath: str
        fpath for the .gpz project file
    """
    with open(fpath, 'rb') as fin:
        f = fin.read()

    profile_num = 1
    while f.find(b'line%d' % profile_num) != -1:
        # Get the header file
        hd_start = f.find(b'line%d.hd' % (profile_num))
        hd_end = f[hd_start:].find(b'PK') + hd_start
        hd_str = str(f[hd_start:hd_end])
        hd_lines = hd_str.split('\\r\\n')
        hd_lines[0] = hd_lines[0][2:]
        hd_lines[-1] = ''

        # Get the 'ini' file
        ini_start = f.find(b'line%d.ini' % (profile_num))
        ini_end = f[ini_start:].find(b'PK') + ini_start
        ini_str = str(f[ini_start:ini_end])
        for i, line in enumerate(ini_str.split('\\r\\n')):
            if i == 0:
                hd_lines.append(line[2:len('line%d.ini' % (profile_num)) + 2])
                hd_lines.append(line[len('line%d.ini' % (profile_num)) + 2:])
            elif i == len(ini_str.split('\\r\\n')) - 1:
                continue
            else:
                hd_lines.append(line)

        # Write to the header file
        with open('LINE' + str(profile_num) + '.HD', 'w') as fout:
            for line in hd_lines:
                fout.write(line + '\n')

        # Get the data file
        dt_start = f.find(b'line%d.dt1' % (profile_num))
        dt_start += len(b'line%d.dt1' % (profile_num))
        dt_end = f[dt_start:].find(b'Lineset') + dt_start
        dt_str = f[dt_start:dt_end]
        # Write to the data file
        with open('LINE' + str(profile_num) + '.DT1', 'wb') as fout:
            fout.write(dt_str)

        profile_num += 1


def read_dt1(fpath, navcrs, body):
    """
    read  Sensors and Software .DT1 data files. 
    modified from gprpy/toolbox/gprIO_DT1.readdt1
    INPUT: 
    fpath           data file name including the .DT1 extension
    navcrs          nav data coordinate reference system
    body            planetary body for coordinate transformation
    OUTPUT:
    rdata           garlic class object

    Load data from a pulse_ekko file."""

    rdata = garlic(fpath)
    rdata.fn = fpath.split("/")[-1][:-4]
    rdata.dtype = "pekko"
    infile_gps = fpath[:-4] + ".GPS"
    if not os.path.isfile(infile_gps):
        infile_gps = fpath[:-4] + ".GP2"
    infile_hd = fpath[:-4] + ".HD"
    headlen = 32
    with open(fpath,"rb") as datafile:
        datafile.seek(8,0) # 0 is beginning of file       
        snum, = struct.unpack('f',datafile.read(4))
        rdata.snum = int(snum)
        dimtrace = rdata.snum*2+128
        datafile.seek(-dimtrace,2) # 2 stands for end of file
        tnum, = struct.unpack('f',datafile.read(4))
        rdata.tnum = int(tnum)
        # Initialize matrix
        rdata.set_dat(np.zeros((rdata.snum,rdata.tnum)))
        head = np.zeros((headlen,rdata.tnum))
        # Set the reader to the beginning of the file
        datafile.seek(0,0)
        for j in range(0,rdata.tnum):
            # Read the header info
            for k in range(0,headlen):
                info, =  struct.unpack('f',datafile.read(4))
                head[k,j] = info
            # Now the actual data
            for k in range(0,rdata.snum):
                # 2 is the size of short, 'h' is the symbol
                pnt, = struct.unpack('h',datafile.read(2))
                rdata.dat[k,j] = pnt
            datafile.seek(dimtrace*(j+1),0) 
    # known vars that are not really set
    rdata.nchan = 1
    rdata.trace_num = np.arange(rdata.tnum) + 1

    # read in header file and return info dict
    rdata.info = read_hd(infile_hd)

    # get sampling frequency
    rdata.dt = rdata.info["Total_time_window"] / rdata.snum * 1.0e-9
    rdata.info.pop("Total_time_window")

    # convert signed int amplitude to floating point for displaying
    rdata.set_proc(rdata.get_dat().astype(float))

    rdata.set_twtt()

    # create nav object to hold lon, lat, elev
    rdata.geocrs = navcrs
    rdata.xyzcrs = navparse.xyzsys[body]
    rdata.navdf = navparse.getnav_pulseekko(infile_gps, rdata.tnum, navcrs, body)

    # for ground-based GPR, elev_gnd is the same as GPS recorded elev
    rdata.set_srfElev(dat = rdata.navdf["elev"])

    rdata.check_attrs()

    return rdata
        
       
def read_hd(fpath):
    '''
    read Sensors and Software .HD header files
    modified from gprpy/toolbox/gprIO_DT1.readdt1Header
    INPUT: 
    fpath           header file name including the .HD extension
    OUTPUT:
    info            dict with information from the header
    '''
    
    info = {}
    with open(fpath,"r",newline='\n') as datafile:
        datafile.readline().strip()
        info["System"] = datafile.readline().strip().split(" ")[0]
        info["Date"] = datafile.readline().strip()
        string = datafile.readline().strip()
        var = re.match(r'NUMBER OF TRACES   = (.*)', string)
        info["N_traces"] = int(var.group(1)) 
        string = datafile.readline().strip()
        var = re.match(r'NUMBER OF PTS/TRC  = (.*)', string)
        info["N_pts_per_trace"] = int(var.group(1)) 
        string = datafile.readline().strip()
        var = re.match(r'TIMEZERO AT POINT  = (.*)', string)
        info["TZ_at_pt"] = float(var.group(1)) 
        string = datafile.readline().strip()
        var = re.match(r'TOTAL TIME WINDOW  = (.*)', string)
        info["Total_time_window"] = float(var.group(1)) 
        string = datafile.readline().strip()
        var = re.match(r'STARTING POSITION  = (.*)', string)
        info["Start Pos."] = float(var.group(1)) 
        string = datafile.readline().strip()
        var = re.match(r'FINAL POSITION     = (.*)', string)
        info["End Pos."] = float(var.group(1))
        string = datafile.readline().strip()
        var = re.match(r'STEP SIZE USED     = (.*)', string)
        info["Step"] = float(var.group(1))
        string = datafile.readline().strip()
        var = re.match(r'POSITION UNITS     = (.*)', string)
        info["Unit"] = str(var.group(1))
        string = datafile.readline().strip()
        var = re.match(r'NOMINAL FREQUENCY  = (.*)', string)
        info["Freq [MHz]"] = float(var.group(1))
        string = datafile.readline().strip()
        var = re.match(r'ANTENNA SEPARATION = (.*)', string)
        info["Antenna Sep."] = float(var.group(1))                   
        # If you need more of the header info, you can just continue as above
        # Transform feet to meters
        if info['Unit'] == 'ft':
            info["Start Pos."] = info["Start_pos"]*0.3048
            info["End Pos."] = info["Final_pos"]*0.3048
            info["Step"] = info["Step_size"]*0.3048
            info["Antenna Sep."] = info["Antenna_sep"]*0.3048
            info['Unit'] = 'm'

        # rearange info dict
        out = {}
        keys = ["System", "Freq [MHz]", "Antenna Sep.", "Start Pos.", "End Pos.", "Step", "Unit", "Total_time_window"]
        for key in keys:
            out[key] = info[key]

    return out

# # method to read pulseekko data
# def read(fpath, navcrs, body):
#     """Load data from a pulse_ekko file."""
#     rdata = garlic(fpath)
#     rdata.fn = fpath.split("/")[-1][:-4]
#     rdata.dtype = "pekko"
#     print("----------------------------------------")
#     print("Loading: " + rdata.fn)
#     infile_gps = fpath[:-4] + ".GPS"
#     infile_hd = fpath[:-4] + ".HD"
#     try:
#         strtypes = (unicode, str)
#         openmode_unicode = 'rU'
#     except NameError:
#         strtypes = (str, )
#         openmode_unicode = 'r'

#     with open(infile_hd, openmode_unicode) as fin:
#         if fin.read().find('1.5.340') != -1:
#             pe_version = '1.5.340'
#         else:
#             pe_version = '1.0'
#         print(pe_version)

#         fin.seek(0)
#         for i, line in enumerate(fin):
#             if 'TRACES' in line or 'NUMBER OF TRACES' in line:
#                 rdata.tnum = int(line.rstrip('\n\r ').split(' ')[-1])
#             if 'PTS' in line or 'NUMBER OF PTS/TRC' in line:
#                 rdata.snum = int(line.rstrip('\n\r ').split(' ')[-1])
#             if ('WINDOW' in line and 'AMPLITUDE' not in line) or 'TOTAL TIME WINDOW' in line:
#                 window = float(line.rstrip('\n\r ').split(' ')[-1])
#             if 'TIMEZERO' in line or 'TIMEZERO AT POINT' in line:
#                 trig = int(float(line.rstrip('\n\r ').split(' ')[-1])
#                                    ) * np.ones((rdata.tnum,))
#             # if i == 4 and pe_version == '1.0':
#             #     try:
#             #         doy = (int(line[6:10]), int(line[1:2]), int(line[3:5]))
#             #     except ValueError:
#             #         doy = (int(line[:4]), int(line[5:7]), int(line[8:10]))
#             # if i == 2 and pe_version == '1.5.340':
#             #     doy = (int(line[6:10]), int(line[:2]), int(line[3:5]))

#     if pe_version == '1.0':
#         rdata.dat = np.zeros((rdata.snum, rdata.tnum), dtype=np.int16)
#     elif pe_version == '1.5.340':
#         rdata.dat = np.zeros((rdata.snum, rdata.tnum), dtype=np.float32)

#     rdata.traceheaders = TraceHeaders(rdata.tnum)
#     with open(fpath, 'rb') as fin:
#         lines = fin.read()

#     offset = 0
#     for i in range(rdata.tnum):
#         # rdata.traceheaders.get_header(offset, lines)
#         offset += 25 * 8 + 28
#         # if pe_version == '1.0':
#         #     trace = struct.unpack('<{:d}h'.format(rdata.snum),
#         #                           lines[offset: offset + rdata.snum * 2])
#         #     offset += rdata.snum * 2
#         # elif pe_version == '1.5.340':
#         #     fmt = '<%df' % (len(lines[offset: offset + rdata.snum * 4]) // 4)
#         #     trace = struct.unpack(fmt, lines[offset:offset + rdata.snum * 4])
#         #     offset += rdata.snum * 4
#         if pe_version == '1.5.340':
#             fmt = '<%df' % (len(lines[offset: offset + rdata.snum * 2]) // 2)
#             trace = struct.unpack(fmt, lines[offset:offset + rdata.snum * 2])
#             offset += rdata.snum * 2

#         # trace -= np.nanmean(trace[:100])
#         # rdata.dat[:, i] = trace.copy()
#         print('here')
#         try:
#             rdata.dat[:,i] = trace
#         except:
#             pass
#         print(trace)
#         print(len(trace))
#         print(rdata.snum)


#     # known vars that are not really set
#     rdata.nchan = 1
#     rdata.trace_num = np.arange(rdata.tnum) + 1

#     # Power some more real variables
#     rdata.dt = window / rdata.snum * 1.0e-9

#     # convert signed int amplitude to floating point for displaying
#     rdata.set_proc(rdata.dat.astype(float))

#     rdata.set_sim(np.ones(rdata.dat.shape))                # place holder for clutter data

#     # assign signal info
#     rdata.info["signal type"] = "impulse"

#     # create nav object to hold lon, lat, elev
#     rdata.navdf = navparse.getnav_pulseekko(infile_gps, rdata.tnum, navcrs, body)

#     # for ground-based GPR, elev_gnd is the same as GPS recorded elev
#     rdata.set_srfElev(rdata.navdf["elev"])

#     return rdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
""" 
constants
"""
### imports ###

# some physical constants for Maxwell's equation for speed of light in a dielectric medium
C = 299792458                   # speed of light in a vacuum
# Eps_0 = 8.8541878 * 10**(-12)   # epsilon naught (vacuum permittivity)
# Mu_0 = 1.257 * 10**(-6)         # mu naught (vacuum permeability)


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
impick class is a tkinter frame which handles the RAGU profile view and radar data picking
"""
### imports ###
from ragu.tools import utils, export
from ragu.ui import basemap
import numpy as np
import tkinter as tk
import sys,os,time,fnmatch,copy,PIL
import matplotlib as mpl
mpl.use("TkAgg")
import matplotlib.pyplot as plt
from matplotlib.widgets import Cursor
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from cycler import cycler
from scipy.interpolate import CubicSpline
# try:
#     plt.rcParams["font.family"] = "Times New Roman"
# except:
#     pass

class impick(tk.Frame):
    # initialize impick frame with variables passed from mainGUI
    def __init__(self, parent, button_tip, popup, fs):
        tk.Frame.__init__(self, parent)
        self.parent = parent
        self.button_tip = button_tip
        self.popup = popup
        # variables only setup once
        self.im_status = tk.IntVar(value=0)
        self.chan = tk.IntVar()
        self.winSize = tk.IntVar(value=0)
        self.horVar = tk.StringVar()
        self.segVar = tk.IntVar()
        self.color = tk.StringVar()
        self.setup(fs)


    # setup impick
    def setup(self, fs):
        # set up frames
        infoFrame0 = tk.Frame(self.parent)
        infoFrame0.pack(side="top",fill="both")        
        infoFrame = tk.Frame(infoFrame0)
        infoFrame.pack(side="left",fill="both")
        interpFrame = tk.Frame(infoFrame0, width=450)
        interpFrame.pack(side="right",fill="both")
        interpFrame.pack_propagate(0)
        toolbarFrame = tk.Frame(infoFrame)
        toolbarFrame.pack(side="bottom",fill="both")
        self.dataFrame = tk.Frame(self.parent)
        self.dataFrame.pack(side="bottom", fill="both", expand=1)

        # add radio buttons for toggling between radargram and clutter sim
        radarRadio = tk.Radiobutton(infoFrame, text="Radargram", variable=self.im_status, value=0, command=self.set_im)
        radarRadio.pack(side="left")
        self.button_tip(self.parent, radarRadio, "View radar profile")

        self.simRadio = tk.Radiobutton(infoFrame,text="Clutter Sim", variable=self.im_status, value=1, command=self.set_im)
        self.simRadio.pack(side="left")
        self.button_tip(self.parent, self.simRadio, "View clutter simulation")
        tk.ttk.Separator(infoFrame,orient="vertical").pack(side="left", fill="both", padx=10, pady=4)

        # add radio buttons for toggling data channel
        tk.Label(infoFrame, text="Channel: ").pack(side="left")
        button = tk.Radiobutton(infoFrame,text="0", variable=self.chan, value=0, command=self.switchChan)
        button.pack(side="left")
        self.button_tip(self.parent, button, text="Radar channel 0")
        button = tk.Radiobutton(infoFrame,text="1", variable=self.chan, value=1, command=self.switchChan)
        button.pack(side="left")
        self.button_tip(self.parent, button, text="Radar channel 1")
        tk.ttk.Separator(infoFrame,orient="vertical").pack(side="left", fill="both", padx=10, pady=4)

        # add entry box for peak finder window size
        tk.Label(infoFrame, text = "Amplitude Window [#Samples]: ").pack(side="left")
        entry = tk.Entry(infoFrame, textvariable=self.winSize, width = 5)
        entry.pack(side="left")
        self.button_tip(self.parent, entry, \
            "Pick amplitude window size. Window size (number of samples) centered on \
            manual picks from which to select maximum amplitude sample for selected trace.")
        tk.ttk.Separator(infoFrame,orient="vertical").pack(side="left", fill="both", padx=10, pady=4)

        # set up frame to hold pick information
        interpFrameT = tk.Frame(interpFrame)
        interpFrameT.pack(fill="both",expand=True)
        interpFrameT.pack_propagate(0)
        interpFrameB = tk.Frame(interpFrame)
        interpFrameB.pack(fill="both",expand=True)
        interpFrameB.pack_propagate(0)

        interpFrameTl = tk.Frame(interpFrameT,width=350,relief="ridge", borderwidth=1)
        interpFrameTl.pack(side="left",fill="both",expand=True)
        interpFrameTl.pack_propagate(0)
        interpFrameTr = tk.Frame(interpFrameT,width=100)
        interpFrameTr.pack(side="left",fill="both",expand=True)
        interpFrameTr.pack_propagate(0)

        interpFrameBl = tk.Frame(interpFrameB,width=350,relief="ridge", borderwidth=1)
        interpFrameBl.pack(side="left",fill="both",expand=True)
        interpFrameBl.pack_propagate(0)
        interpFrameBr = tk.Frame(interpFrameB,width=100)
        interpFrameBr.pack(side="left",fill="both",expand=True)
        interpFrameBr.pack_propagate(0)

        tk.Label(interpFrameTl,text="{0:<8}".format("Horizon:")).pack(side="left")
        button = tk.Button(interpFrameTl, text="Delete", width=4, command=lambda:self.rm_horizon(horizon=self.horVar.get()))
        button.pack(side="right")
        self.button_tip(self.parent, button, "Remove interpretation horion")
        button = tk.Button(interpFrameTl, text="Rename", width=4, command=lambda:self.rename_horizon(horizon=self.horVar.get()))
        button.pack(side="right")
        self.button_tip(self.parent, button, "Rename interpretation horion")
        button = tk.Button(interpFrameTl, text="New", width=4, command=self.init_horizon)
        button.pack(side="right")
        self.button_tip(self.parent, button, "New interpretation horion")
        self.horMenu = tk.OptionMenu(interpFrameTl, self.horVar, *[None])
        self.horMenu.pack(side="right")
        self.horMenu.config(width=10)
        self.horVar.trace("w", lambda *args, last=True : self.update_seg_opt_menu(last)) 
        self.horVar.trace("w", lambda *args, menu=self.horMenu, var="horVar" : self.set_menu_color(menu, var))

        tk.Label(interpFrameBl,text="{0:<8}".format("Segment:")).pack(side="left")
        button = tk.Button(interpFrameBl, text="Delete", width=4, command=lambda:self.rm_segment(horizon=self.horVar.get(),seg=self.segVar.get()))
        button.pack(side="right")
        self.button_tip(self.parent, button, "Remove interpretation segment")
        button = tk.Button(interpFrameBl, text="Edit", width=4, command=lambda:self.edit_segment(horizon=self.horVar.get(),seg=self.segVar.get()))
        button.pack(side="right")
        self.button_tip(self.parent, button, "Edit interpretation segment")    
        button = tk.Button(interpFrameBl, text="New", width=4, command=lambda:self.init_segment(horizon=self.horVar.get()))
        button.pack(side="right")
        self.button_tip(self.parent, button, "New interpretation segment")
        self.segMenu = tk.OptionMenu(interpFrameBl, self.segVar, *[None])
        self.segMenu.pack(side="right")
        self.segMenu.config(width=10)
    
        # initialize pick state buttons
        label = tk.Label(interpFrameTr, text="Pick",justify="center")
        label.pack(fill="both",expand=True)
        f = tk.font.Font(label, label.cget("font"))
        f.configure(underline=True)
        label.configure(font=f)
        self.startbutton = tk.Button(interpFrameBr, text="Start", bg="green", fg="white", command=lambda:self.set_pickState(state=True))
        self.startbutton.pack(side="left",fill="both",expand=True)
        self.button_tip(self.parent, self.startbutton, "Start picking")
        self.stopbutton = tk.Button(interpFrameBr, text="Stop", bg="red", fg="white", command=lambda:self.set_pickState(state=False))
        self.stopbutton.pack(side="left",fill="both",expand=True)
        self.button_tip(self.parent, self.stopbutton, "Stop picking")

        # create matplotlib figure data canvas
        plt.rcParams.update({'font.size': fs, 'axes.titlesize': fs})
        self.fig = mpl.figure.Figure()
        self.fig.patch.set_facecolor("#d9d9d9")
        self.dataCanvas = FigureCanvasTkAgg(self.fig, self.parent)
        self.ax = self.fig.add_subplot(1,1,1)
        self.fig.tight_layout(rect=[.05,.05,.97,1])

        # initiate a twin axis that shows twtt
        self.secaxy0 = self.ax.twinx()
        self.secaxy0.yaxis.set_ticks_position("right")
        self.secaxy0.yaxis.set_label_position("right")

        # initiate a twin axis that shares the same x-axis and shows approximate depth
        self.secaxy1 = self.ax.twinx()
        self.secaxy1.yaxis.set_ticks_position("left")
        self.secaxy1.yaxis.set_label_position("left")
        self.secaxy1.spines["left"].set_position(("outward", 50))

        # initiate a twin axis that shows along-track distance
        self.secaxx = self.ax.twiny()
        self.secaxx.xaxis.set_ticks_position("bottom")
        self.secaxx.xaxis.set_label_position("bottom")
        self.secaxx.spines["bottom"].set_position(("outward", 35))

        # set zorder of secondary axes to be behind main axis (self.ax)
        self.secaxx.set_zorder(-100)
        self.secaxy0.set_zorder(-100)
        self.secaxy1.set_zorder(-100)

        # add axes for colormap sliders and reset button - leave invisible until rdata loaded
        self.ax_cmax = self.fig.add_axes([0.97, 0.55, 0.0075, 0.30])
        self.ax_cmin  = self.fig.add_axes([0.97, 0.18, 0.0075, 0.30])
        self.reset_ax = self.fig.add_axes([0.95625, 0.10, 0.03, 0.035])
     
        # create colormap sliders and reset button - initialize for data image
        self.s_cmin = mpl.widgets.Slider(self.ax_cmin, "Min", 0, 1, orientation="vertical", handle_style={"size":fs})
        self.s_cmax = mpl.widgets.Slider(self.ax_cmax, "Max", 0, 1, orientation="vertical", handle_style={"size":fs})
        self.cmap_reset_button = mpl.widgets.Button(self.reset_ax, "Reset", color="white")
        self.cmap_reset_button.label.set_fontsize(fs)
        self.cmap_reset_button.on_clicked(self.cmap_reset)

        # add toolbar to plot
        self.toolbar = NavigationToolbar2Tk(self.dataCanvas, toolbarFrame)
        self.toolbar.update()

        # canvas callbacks
        self.click = self.fig.canvas.mpl_connect("button_press_event", self.onpress)
        self.unclick = self.fig.canvas.mpl_connect("button_release_event", self.onrelease)
        self.draw_cid = self.fig.canvas.mpl_connect("draw_event", self.update_bg)
        self.resize_cid = self.fig.canvas.mpl_connect("resize_event", self.drawData)
        self.mousemotion = self.fig.canvas.mpl_connect('motion_notify_event', self.on_mouse_move)

        # set custom mpl line colors - cyan, green, orange, magenta, yellow, blue, purple, brown
        self.ln_colors = {}
        self.ln_colors["str"] = ["cyan", "green", "orange", "magenta", "yellow", "blue", "purple", "brown"]
        self.ln_colors["hex"] = ["#17becf","#2ca02c","#ff7f0e","#e377c2","#bcbd22","#1f77bf","#9467bd","#8c564b"]
        mpl.rcParams['axes.prop_cycle'] = cycler(color=self.ln_colors["hex"])


    # set_vars is a method to set impick variables which need to reset upon each load
    def set_vars(self):
        self.pick_vis = True
        self.ann_vis = True
        self.basemap = None
        self.pick_surf = None

        self.pick_state = False
        self.pyramid = None

        # image colormap bounds
        self.data_cmin = None
        self.data_cmax = None
        self.sim_cmin = None
        self.sim_cmax = None

        # image colormap range
        self.data_crange = None
        self.sim_crange = None

        # initialize path objects #
        self.tmp_horizon_path = path([],[])                                     # temporary path object to hold horizon segment currently being picked
        self.edit_path = None                                                   # temporary path object to hold horizon segment currently being edited
        self.horizon_paths = {}                                                 # dictionary to hold horizon paths for saved picks

        # initialize line objects
        self.tmp_horizon_ln = None
        self.horizon_lns = {}
        self.horizontal_line = None
        self.vertical_line = None

        # initialize list of pick annotations
        self.ann_list = []
        self.horizons = []

        # necessary flags
        self.sim_imSwitch_flag = False
        self.edit_flag = False

        self.im_status.set(0)
        self.debugState = False
        self.horVar.set("")
        self.segVar.set(0)
        self.color.set("")
        self.ln_colors["used"] = {}


    # set image cmap
    def set_cmap(self, cmap):
        cmap = copy.copy(mpl.cm.get_cmap(cmap))
        self.cmap = mpl.cm.get_cmap(cmap)
        # set nodata value as bottom color in cmap
        nd = self.cmap(np.linspace(0,1,256))[0]
        self.cmap.set_bad(nd)


    # load calls ingest() on the data file and sets the datacanvas
    def load(self, rdata):       
        # receive the rdata
        self.rdata = rdata

        # pack the datacanvas in data frame
        self.dataCanvas.get_tk_widget().pack(in_=self.dataFrame, side="bottom", fill="both", expand=1) 
      
        # set figure title and axes labels
        self.ax.set_title(self.rdata.fn)
        self.ax.set(xlabel = "Trace", ylabel = "Sample")

        # initialize data and clutter images with np.ones - set data in drawData
        self.im_dat  = self.ax.imshow(np.ones((100,100)), aspect="auto", 
                        extent=[0, self.rdata.tnum, self.rdata.snum, 0])
        self.im_sim  = self.ax.imshow(np.ones((100,100)), aspect="auto", 
                        extent=[0, self.rdata.tnum, self.rdata.snum, 0])

        # set clutter sim visibility to false
        self.im_sim.set_visible(False)
        # disable im toggle if no sim
        if not self.rdata.flags.sim:
            self.simRadio.config(state="disabled")

        # initialize line to hold current picks
        self.tmp_horizon_ln, = self.ax.plot(self.tmp_horizon_path.x, self.tmp_horizon_path.y, "rx", ms=6)
        # initialize cursor crosshair lines
        self.horizontal_line = self.ax.axhline(color="r", lw=1, ls="--")
        self.vertical_line = self.ax.axvline(color="r", lw=1, ls="--")
        self.set_cross_hair_visible(self.get_pickState())

        # update the canvas
        self.dataCanvas._tkcanvas.pack()

        # connect ylim_change with event to update image pyramiding based on zoom - have to do this on load, since clear_canvas removes axis callbacks
        self.ylim_cid = self.ax.callbacks.connect("ylim_changed", self.drawData)

        # update toolbar to save axes extents
        self.toolbar.update()

        # set color range
        self.set_crange()


    # set radar and sim array bounds for setting image color limits - just doing this once based off original arrays, not pyramids
    def set_crange(self):
        # get clim bounds - take 10th percentile for min, ignore nd values
        self.mindB_data = np.nanpercentile(self.rdata.proc.get_curr_dB(),10)
        self.maxdB_data = np.nanpercentile(self.rdata.proc.get_curr_dB(),99)
        # self.maxdB_data = np.nanmax(self.rdata.proc.get_curr_dB())

        if self.rdata.flags.sim:
            self.mindB_sim = np.floor(np.nanpercentile(self.rdata.sim,10))
            self.maxdB_sim = np.nanmax(self.rdata.sim)
            self.sim_crange = self.maxdB_sim - self.mindB_sim
            self.im_sim.set_clim([self.mindB_sim, self.maxdB_sim])

        # get colormap range
        self.data_crange = self.maxdB_data - self.mindB_data
        # update color limits
        self.im_dat.set_clim([self.mindB_data, self.maxdB_data])

        # set slider bounds - use data clim values upon initial load
        self.s_cmin.valmin = self.mindB_data - (self.data_crange/2)
        self.s_cmin.valmax = self.mindB_data + (self.data_crange/2)
        self.s_cmin.valinit = self.mindB_data
        self.s_cmax.valmin = self.maxdB_data - (self.data_crange/2)
        self.s_cmax.valmax = self.maxdB_data + (self.data_crange/2)
        self.s_cmax.valinit = self.maxdB_data
        self.update_slider()


    # update clim slider bar
    def update_slider(self):
        self.ax_cmax.clear()
        self.ax_cmin.clear()
        self.s_cmin.__init__(self.ax_cmin, "Min", valmin=self.s_cmin.valmin, valmax=self.s_cmin.valmax, valinit=self.s_cmin.valinit, orientation="vertical")
        self.s_cmax.__init__(self.ax_cmax, "Max", valmin=self.s_cmax.valmin, valmax=self.s_cmax.valmax, valinit=self.s_cmax.valinit, orientation="vertical")
        self.s_cmin.on_changed(self.cmap_update)
        self.s_cmax.on_changed(self.cmap_update)


    # update image colormap based on slider values
    def cmap_update(self, s=None):
        try:
            if self.im_dat.get_visible():
                # apply slider values to visible image
                self.data_cmin = self.s_cmin.val
                self.data_cmax = self.s_cmax.val
                self.im_dat.set_clim([self.data_cmin, self.data_cmax])
                # update norm
                # self.im_dat.set_norm(mpl.colors.LogNorm(self.data_cmin, self.data_cmax))
            else:
                self.sim_cmin = self.s_cmin.val
                self.sim_cmax = self.s_cmax.val
                self.im_sim.set_clim([self.sim_cmin, self.sim_cmax])
                # update norm
                # self.im_sim.set_norm(mpl.colors.LogNorm(self.sim_cmin, self.sim_cmax))
        except Exception as err:
            print("cmap_update error: " + str(err))


    # reset sliders to initial values
    def cmap_reset(self, event):
        if self.im_dat.get_visible():
            self.s_cmin.valmin = self.mindB_data - (self.data_crange/2)
            self.s_cmin.valmax = self.mindB_data + (self.data_crange/2)
            self.s_cmin.valinit = self.mindB_data
            self.s_cmax.valmin = self.maxdB_data - (self.data_crange/2)
            self.s_cmax.valmax = self.maxdB_data + (self.data_crange/2)
            self.s_cmax.valinit = self.maxdB_data
        else:
            # if sim is displayed, change slider bounds
            self.s_cmin.valmin = self.mindB_sim - (self.sim_crange/2)
            self.s_cmin.valmax = self.mindB_sim + (self.sim_crange/2)
            self.s_cmin.valinit = self.mindB_sim
            self.s_cmax.valmin = self.maxdB_sim - (self.sim_crange/2)
            self.s_cmax.valmax = self.maxdB_sim + (self.sim_crange/2)
            self.s_cmax.valinit = self.maxdB_sim
        self.update_slider()
        self.cmap_update()


    # method to draw radar data
    def drawData(self, force=False, event=None):
        # Get data display window size in inches
        w,h = self.fig.get_size_inches()*self.fig.dpi
        # choose pyramid
        p = -1
        for i in range(len(self.rdata.dPyramid)-1, -1, -1):
            if(self.rdata.dPyramid[i].shape[0] > h):
                p = i
                break

        # set flag to detect if canvas needs redrawing
        flag = False

        # if ideal pyramid level changed, update image
        if self.pyramid != p or force:
            self.pyramid = p
            if len(self.rdata.dPyramid[self.pyramid].shape) == 3:
                self.im_dat.set_data(self.rdata.dPyramid[self.pyramid][:,:,self.chan.get()])
            else:
                self.im_dat.set_data(self.rdata.dPyramid[self.pyramid][:,:])
            if self.rdata.flags.sim:
                self.im_sim.set_data(self.rdata.sPyramid[self.pyramid][:,:])
            flag = True

        # update cmap if necessary
        if self.im_dat.get_cmap().name != self.cmap.name:
            self.im_dat.set_cmap(self.cmap)
            self.im_sim.set_cmap(self.cmap)
            flag = True

        if flag:
            self.dataCanvas.draw()


    # set axis labels
    def set_axes(self):
        # update twtt and depth (subradar dist.)
        if self.rdata.dt < 1e-9:
            self.secaxy0.set_ylabel("TWTT [ns]")
            self.secaxy0.set_ylim(self.rdata.snum * self.rdata.dt * 1e9, 0)
        else:
            self.secaxy0.set_ylabel("TWTT [\u03BCs]")
            self.secaxy0.set_ylim(self.rdata.snum * self.rdata.dt * 1e6, 0)

        self.secaxy1.set_ylabel("Depth [m] (" + r"$\epsilon_{r}$" + f"= {self.eps_r})")
        self.secaxy1.set_ylim(utils.twtt2depth(self.rdata.snum * self.rdata.dt, np.nanmean(self.rdata.asep), self.eps_r), 0)

        # update along-track distance
        if not np.all(np.isnan(self.rdata.navdf["dist"])) or  np.all((self.rdata.navdf["dist"] == 0)):
            self.secaxx.set_visible(True)
            # use km if distance exceeds 1 km
            if self.rdata.navdf.iloc[-1]["dist"] >= 1e3:
                self.secaxx.set_xlabel("Along-Track Distance [km]")
                self.secaxx.set_xlim(0, self.rdata.navdf.iloc[-1]["dist"]*1e-3)

            else:
                self.secaxx.set_xlabel("Along-Track Distance [m]")
                self.secaxx.set_xlim(0, self.rdata.navdf.iloc[-1]["dist"])
        
        else:
            self.secaxx.set_visible(False)


    # method to zoom to full image extent
    def fullExtent(self):
        self.ax.set_xlim(0, self.rdata.tnum)
        self.ax.set_ylim(self.rdata.snum, 0)
        self.set_axes()
        self.dataCanvas.draw()


    # method to vertically clip rgam for export
    def verticalClip(self, top=0.0, bottom = 0.5):
        self.ax.set_ylim(self.rdata.snum*bottom, self.rdata.snum*top)
        ylim2 = self.secaxy0.get_ylim()
        ylim3 = self.secaxy1.get_ylim()
        self.secaxy0.set_ylim(ylim2[0]*bottom,ylim2[0]*top)
        self.secaxy1.set_ylim(ylim3[0]*bottom,ylim3[0]*top)
        self.dataCanvas.draw()

    # method to zoom in
    def zoomIn(self, factor=1/4):
        # get original axis limits
        xlima1 = self.ax.get_xlim()
        xlima2 = self.secaxx.get_xlim()
        ylima1 = self.ax.get_ylim()
        ylima2 = self.secaxy0.get_ylim()
        ylima3 = self.secaxy1.get_ylim()

        xstep1 = abs(xlima1[1] - xlima1[0]) * factor
        xstep2 = abs(xlima2[1] - xlima2[0]) * factor
        ystep1 = abs(ylima1[1] - ylima1[0]) * factor
        ystep2 = abs(ylima2[1] - ylima2[0]) * factor
        ystep3 = abs(ylima3[1] - ylima3[0]) * factor

        # add steps
        xlimb1 = [xlima1[0]+xstep1, xlima1[1]-xstep1]
        xlimb2 = [xlima2[0]+xstep2, xlima2[1]-xstep2]
        ylimb1 = [ylima1[0]-ystep1, ylima1[1]+ystep1]
        ylimb2 = [ylima2[0]-ystep2, ylima2[1]+ystep2]
        ylimb3 = [ylima3[0]-ystep3, ylima3[1]+ystep3]

        self.ax.set_xlim(xlimb1[0], xlimb1[1])
        self.secaxx.set_xlim(xlimb2[0], xlimb2[1])
        self.ax.set_ylim(ylimb1[0], ylimb1[1])
        self.secaxy0.set_ylim(ylimb2[0], ylimb2[1])
        self.secaxy1.set_ylim(ylimb3[0], ylimb3[1])
        self.dataCanvas.draw()


    # method to zoom out
    def zoomOut(self, factor=1/4):
        # get original axis limits
        xlima1 = self.ax.get_xlim()
        xlima2 = self.secaxx.get_xlim()
        ylima1 = self.ax.get_ylim()
        ylima2 = self.secaxy0.get_ylim()
        ylima3 = self.secaxy1.get_ylim()

        xstep1 = abs(xlima1[1] - xlima1[0]) * factor
        xstep2 = abs(xlima2[1] - xlima2[0]) * factor
        ystep1 = abs(ylima1[1] - ylima1[0]) * factor
        ystep2 = abs(ylima2[1] - ylima2[0]) * factor
        ystep3 = abs(ylima3[1] - ylima3[0]) * factor

        # add steps
        xlimb1 = [xlima1[0]-xstep1, xlima1[1]+xstep1]
        xlimb2 = [xlima2[0]-xstep2, xlima2[1]+xstep2]
        ylimb1 = [ylima1[0]+ystep1, ylima1[1]-ystep1]
        ylimb2 = [ylima2[0]+ystep2, ylima2[1]-ystep2]
        ylimb3 = [ylima3[0]+ystep3, ylima3[1]-ystep3]
        # handle bounds
        if xlimb1[0] < 0:
            xlimb1[0] = 0
            xlimb2[0] = 0
        if xlimb1[1] > self.rdata.tnum:
            xlimb1[1] = self.rdata.tnum
            xlimb2[1] = self.rdata.navdf["dist"].iloc[-1]*1e-3
        if ylimb1[0] > self.rdata.snum:
            ylimb1[0] = self.rdata.snum
            ylimb2[0] = self.rdata.snum*self.rdata.dt
            ylimb3[0] = utils.twtt2depth(self.rdata.snum*self.rdata.dt, np.nanmean(self.rdata.asep), self.eps_r)
        if ylimb1[1] < 0:
            ylimb1[1] = 0
            ylimb2[1] = 0
            ylimb3[1] = 0
        self.ax.set_xlim(xlimb1[0], xlimb1[1])
        self.secaxx.set_xlim(xlimb2[0], xlimb2[1])
        self.ax.set_ylim(ylimb1[0], ylimb1[1])
        self.secaxy0.set_ylim(ylimb2[0], ylimb2[1])
        self.secaxy1.set_ylim(ylimb3[0], ylimb3[1])
        self.dataCanvas.draw()


    # method to pan right
    def panRight(self):
        xlim1 = self.ax.get_xlim()
        xlim2 = self.secaxx.get_xlim()
        if xlim1[1] < self.rdata.tnum:
            step1 = (xlim1[1] - xlim1[0]) / 2
            step2 = (xlim2[1] - xlim2[0]) / 2
            if (xlim1[1] + step1) > self.rdata.tnum:
                step1 = self.rdata.tnum - xlim1[1]
                step2 = self.rdata.navdf["dist"].iloc[-1]*1e-3 - xlim2[1]
            self.ax.set_xlim(xlim1[0] + step1, xlim1[1] + step1)
            self.secaxx.set_xlim(xlim2[0] + step2, xlim2[1] + step2)
            self.dataCanvas.draw()


    # method to pan left
    def panLeft(self):
        xlim1 = self.ax.get_xlim()
        xlim2 = self.secaxx.get_xlim()
        if xlim1[0] > 0:
            step1 = (xlim1[1] - xlim1[0]) / 2
            step2 = (xlim2[1] - xlim2[0]) / 2
            if (xlim1[0] - step1) < 0:
                step1 = xlim1[0]
                step2 = xlim2[0]
            self.ax.set_xlim(xlim1[0] - step1, xlim1[1] - step1)
            self.secaxx.set_xlim(xlim2[0] - step2, xlim2[1] - step2)
            self.dataCanvas.draw()


    # method to pan up
    def panUp(self):
        ylim1 = self.ax.get_ylim()
        ylim2 = self.secaxy0.get_ylim()
        ylim3 = self.secaxy1.get_ylim()
        if ylim1[1] > 0:
            step1 = (ylim1[0] - ylim1[1]) / 2
            step2 = (ylim2[0] - ylim2[1]) / 2
            step3 = (ylim3[0] - ylim3[1]) / 2
            if (ylim1[1] - step1) < 0:
                step1 = ylim1[1]
                step2 = ylim2[1]
                step3 = ylim3[1]
            self.ax.set_ylim(ylim1[0] - step1, ylim1[1] - step1)
            self.secaxy0.set_ylim(ylim1[0] - step2, ylim2[1] - step2)
            self.secaxy1.set_ylim(ylim3[0] - step3, ylim3[1] - step3)
            self.dataCanvas.draw()


    # method to pan down
    def panDown(self):
        ylim1 = self.ax.get_ylim()
        ylim2 = self.secaxy0.get_ylim()
        ylim3 = self.secaxy1.get_ylim()
        if ylim1[0] < self.rdata.snum:
            step1 = (ylim1[0] - ylim1[1]) / 2
            step2 = (ylim2[0] - ylim2[1]) / 2
            step3 = (ylim3[0] - ylim3[1]) / 2
            if (ylim1[0] + step1) > self.rdata.snum:
                step1 = self.rdata.snum - ylim1[0]
                step2 = (self.rdata.snum*self.rdata.dt) - ylim2[0]
                step3 = utils.twtt2depth(self.rdata.snum*self.rdata.dt, np.nanmean(self.rdata.asep), self.eps_r) - ylim3[0]
            self.ax.set_ylim(ylim1[0] + step1, ylim1[1] + step1)
            self.secaxy0.set_ylim(ylim1[0] + step2, ylim2[1] + step2)
            self.secaxy1.set_ylim(ylim3[0] + step3, ylim3[1] + step3)
            self.dataCanvas.draw()


    def switchChan(self):
        # check to make sure channel exists before switching
        if self.chan.get() == 1 and self.rdata.nchan < 2:
            self.chan.set(0)
            return
        elif self.chan.get() == 1 and self.rdata.nchan == 2:
            self.drawData(force=True)
        elif self.chan.get() == 0:
            self.drawData(force=True)


   # set_im is a method to set which rdata is being displayed
    def set_im(self, from_gui=False):
        if from_gui and self.rdata.flags.sim:
            self.im_status.set(int(not self.im_status.get()))

        if self.im_status.get() == 0:
            self.show_data()

        elif self.im_status.get() == 1:
            self.show_sim()


    # toggle to radar data
    def show_data(self):
        if not self.im_dat.get_visible():
            # get sim colormap slider values for reviewing
            self.sim_cmin = self.s_cmin.val
            self.sim_cmax = self.s_cmax.val
            # set colorbar initial values to previous values
            self.s_cmin.valinit = self.data_cmin
            self.s_cmax.valinit = self.data_cmax
            # set colorbar bounds
            self.s_cmin.valmin = self.mindB_data - (self.data_crange/2)
            self.s_cmin.valmax = self.mindB_data + (self.data_crange/2)
            self.s_cmax.valmin = self.maxdB_data - (self.data_crange/2)
            self.s_cmax.valmax = self.maxdB_data + (self.data_crange/2)
            self.update_slider()
            # reverse visilibilty
            self.im_sim.set_visible(False)
            self.im_dat.set_visible(True)
            # redraw canvas
            # self.update_bg()
            self.fig.canvas.draw()


    # toggle to sim viewing
    def show_sim(self):
        if not self.im_sim.get_visible():
            # get radar data colormap slider values for reviewing
            self.data_cmin = self.s_cmin.val
            self.data_cmax = self.s_cmax.val

            if not self.sim_imSwitch_flag:
                # if this is the first time viewing the sim, set colorbar limits to initial values
                self.s_cmin.valinit = self.mindB_sim
                self.s_cmax.valinit = self.maxdB_sim
            else: 
                # if sim has been shown before revert to previous colorbar values
                self.im_sim.set_clim([self.sim_cmin, self.sim_cmax])
                self.s_cmin.valinit = self.sim_cmin
                self.s_cmax.valinit = self.sim_cmax

            self.s_cmin.valmin = self.mindB_sim - (self.sim_crange/2)
            self.s_cmin.valmax = self.mindB_sim + (self.sim_crange/2)           
            self.s_cmax.valmin = self.maxdB_sim - (self.sim_crange/2)
            self.s_cmax.valmax = self.maxdB_sim + (self.sim_crange/2)
            self.update_slider()
            # reverse visilibilty
            self.im_dat.set_visible(False)
            self.im_sim.set_visible(True)
            # set flag to indicate that sim has been viewed for resetting colorbar limits
            self.sim_imSwitch_flag = True    
            # redraw canvas
            # self.update_bg()
            self.fig.canvas.draw()


    # get_pickState is a method to return the current picking state
    def get_pickState(self):
        return self.pick_state


    # return surface being picked (surface or subsurface)
    def get_pickSurf(self):
        return self.pick_surf


    # return horizon_paths
    def get_horizon_paths(self):
        return self.horizon_paths

    
    # receive horizon_paths
    def set_horizon_paths(self, horizon_paths):
        self.horizon_paths = copy.deepcopy(horizon_paths)
        for horizon, hdict in self.horizon_paths.items():
            x,y = utils.merge_paths(hdict)
            self.horizon_lns[horizon].set_data(x,y)


    # reverse horizon path objects
    def reverse(self):
        # horizon arrays have already been flipped within proc.reverse, but we need to reset the canvas by redrawing the horizons
        tmp = copy.deepcopy(self.rdata.pick.horizons)
        self.rm_horizon(rm_all=True, verify=False)
        for h in tmp.keys():
            self.rdata.pick.horizons[h] = tmp[h]
            self.set_picks(h)

        self.update_pickLabels()
        self.update_seg_opt_menu()
        self.update_bg()


    # return horizon colors
    def get_horizon_colors(self):
        return self.ln_colors["used"]

    
    # set tkinter menu font colors to match color name
    def set_menu_color(self, menu=None, var=None, *args):
        if menu is None:
            return
        if var=="color":
            c = self.ln_colors["hex"][self.ln_colors["str"].index(self.color.get())]
        elif var=="horVar":
            horizon = self.horVar.get()
            if not horizon:
                return
            c = self.ln_colors["used"][horizon]
        else:
            try:
                c = self.ln_colors["used"][var.get()]
            except:
                return
        menu.config(foreground=c, activeforeground=c, highlightcolor=c)


    # init_horizon is a method to initialize new horizon objects
    def init_horizon(self, horizon=None, skip_array=False):
        if self.get_pickState():
            return
        if not horizon:
            if self.popup.flag == 1:
                return            
            # create popup window to get new horizon name and interpreatation color
            hname = tk.StringVar()
            popup = self.popup.new(title="New Horizon")
            tk.Label(popup, text="Enter new horizon name:").pack(fill="both",expand=True)
            entry = tk.Entry(popup, textvar=hname, justify="center")
            entry.pack(fill="both",expand=True) 
            entry.focus_set()
            # interpretation color selection dropdown - default to first unused color in self.ln_colors
            colors = [c for c in self.ln_colors["hex"] if c not in list(self.ln_colors["used"].values())]
            self.color.set(self.ln_colors["str"][self.ln_colors["hex"].index(colors[0])])
            tk.Label(popup, text="Interpretation color:").pack(fill="both", expand=True)
            dropdown = tk.OptionMenu(popup, self.color, *[None])
            dropdown["menu"].delete(0, "end")
            dropdown.config(width=20)
            dropdown.pack(fill="none", expand=True)                      
            self.set_menu_color(menu=dropdown, var="color")
            # trace change in self.color to self.set_menu_color
            trace = self.color.trace("w", lambda *args, menu=dropdown, var="color" : self.set_menu_color(menu, var))
            for c, h in zip(self.ln_colors["str"], self.ln_colors["hex"]):
                dropdown["menu"].add_command(label=c, foreground=h, activeforeground=h, command=tk._setit(self.color, c))
            button = tk.Button(popup, text="OK", command=lambda:self.popup.close(flag=0), width=20).pack(side="left", fill="none", expand=True)
            button = tk.Button(popup, text="Cancel", command=lambda:[hname.set(""), self.popup.close(flag=-1)], width=20).pack(side="left", fill="none", expand=True)
            # wait for window to be closed
            self.parent.wait_window(popup)
            # remove the trace
            self.color.trace_vdelete("w", trace)
            horizon = hname.get()
            if (not horizon) or (self.popup.flag==-1):
                return

        if not skip_array:
            # ensure horizon doesn't already exist, otherwise overwrite or return
            if horizon not in self.horizon_paths:
                self.rdata.pick.horizons[horizon] = np.repeat(np.nan, self.rdata.tnum)
            elif tk.messagebox.askyesno("Warning","Horizon name (" + horizon + ") already exists. Overwrite?") == True:
                self.rdata.pick.horizons[horizon] = np.repeat(np.nan, self.rdata.tnum)
            else:
                return
        self.horizon_paths[horizon] = {}
        # initialize 0th segment for new horizon
        self.init_segment(horizon=horizon)
        # initialize line object for new horizon
        x,y = utils.merge_paths(self.horizon_paths[horizon])
        self.horizon_lns[horizon], = self.ax.plot(x,y,lw=2,c=self.ln_colors["hex"][self.ln_colors["str"].index(self.color.get())])            
        # update horizon and segment options
        self.update_hor_opt_menu()  
        # set horVar to new horizon
        self.horVar.set(horizon)
        # update crosshairs and tmp pick line colors to match horizon color
        for ln in [self.tmp_horizon_ln, self.horizontal_line, self.vertical_line]:
            ln.set_color(self.ln_colors["str"][self.ln_colors["hex"].index(self.ln_colors["used"][horizon])])


    # rename horizon
    def rename_horizon(self, horizon=None):
        if self.get_pickState():
            return
        if horizon:
            if self.popup.flag == 1:
                return
            # create popup window to rename horizon
            hname = tk.StringVar()
            popup = self.popup.new(title="Rename Horizon")
            tk.Label(popup, text="Rename " + horizon + " horizon:").pack(fill="both",expand=True)
            entry = tk.Entry(popup, textvar=hname, justify="center")
            entry.pack(fill="both",expand=True) 
            entry.focus_set()
            # interpretation color selection dropdown - default to current horizon line color
            self.color.set(self.ln_colors["str"][self.ln_colors["hex"].index(self.ln_colors["used"][horizon])])
            tk.Label(popup, text="Interpretation color:").pack(fill="both", expand=True)
            dropdown = tk.OptionMenu(popup, self.color, *[None])
            dropdown["menu"].delete(0, "end")
            dropdown.config(width=20)
            dropdown.pack(fill="none", expand=True)
            self.set_menu_color(menu=dropdown, var="color")
            # trace change in self.color to self.set_menu_color
            trace = self.color.trace("w", lambda *args, menu=dropdown, var="color" : self.set_menu_color(menu, var))
            for c,h in zip(self.ln_colors["str"], self.ln_colors["hex"]):
                dropdown["menu"].add_command(label=c, foreground=h, activeforeground=h, command=tk._setit(self.color, c))
            button = tk.Button(popup, text="OK", command=lambda:self.popup.close(flag=0), width=20).pack(side="left", fill="none", expand=True)
            button = tk.Button(popup, text="Cancel", command=lambda:[hname.set(""), self.popup.close(flag=-1)], width=20).pack(side="left", fill="none", expand=True)
            # wait for window to be closed
            self.parent.wait_window(popup)
            # remove color trace
            self.color.trace_vdelete("w", trace)
            hname = hname.get()
            if self.popup.flag == -1:
                return
        
            # update line color
            c = self.ln_colors["hex"][self.ln_colors["str"].index(self.color.get())]
            self.ln_colors["used"][horizon] = c
            self.horizon_lns[horizon].set_color(c)
            self.horVar.set(horizon)

            # rename horizon path, line objects, and used line colors maintaining order
            if hname:
                self.horizon_paths = {hname if k==horizon else k:v for k,v in self.horizon_paths.items()}
                self.rdata.pick.horizons = {hname if k==horizon else k:v for k,v in self.rdata.pick.horizons.items()}
                self.horizon_lns = {hname if k==horizon else k:v for k,v in self.horizon_lns.items()}
                self.ln_colors["used"] = {hname if k==horizon else k:v for k,v in self.ln_colors["used"].items()}
                # set horVar to new horizon
                self.horVar.set(hname) 
                self.update_pickLabels()
            # update horizon and segment options
            self.update_hor_opt_menu()  
            self.update_bg()


    # remove horizon
    def rm_horizon(self, rm_all=False, horizon=None, verify=True):
        if len(self.horizons) < 1:
            return

        if rm_all:
            horizon = list(self.horizon_paths.keys())
            if verify:
                if tk.messagebox.askyesno("Warning","Remove all interpretation horizons?"):
                    verify = False
                else:
                    return

        if horizon is None:
            if self.popup.flag == 1:
                return
            # popup window to get horizon and segment option
            hname = tk.StringVar()
            hname.set(self.horVar.get())
            # create popup window to rename horizon
            popup = self.popup.new(title="Remove Horizon")
            # create horizon dropdown menu
            tk.Label(popup, text="Horizon:").pack(fill="both", expand=True)
            hdropdown = tk.OptionMenu(popup, hname, *[None])
            hdropdown.config(width=20)
            hdropdown.pack(fill="none", expand=True)
            # horizon dropdown with ln colors
            self.color.set(self.ln_colors["str"][self.ln_colors["hex"].index(self.ln_colors["used"][hname.get()])])
            self.set_menu_color(menu=hdropdown, var=hname)
            # trace change in self.color to self.set_menu_color
            trace0 = hname.trace("w", lambda *args, menu=hdropdown, var=hname : self.set_menu_color(menu, var))
            for key, val in self.ln_colors["used"].items():
                hdropdown["menu"].add_command(label=key, foreground=val, activeforeground=val, command=tk._setit(hname, key))
            hname.set(self.horVar.get())
            button = tk.Button(popup, text="OK", command=lambda:self.popup.close(flag=0), width=20).pack(side="left", fill="none", expand=True)
            button = tk.Button(popup, text="Cancel", command=lambda:[hname.set(""), self.popup.close(flag=-1)], width=20).pack(side="left", fill="none", expand=True)
            # wait for window to be closed
            self.parent.wait_window(popup)
            # remove color trace
            hname.trace_vdelete("w", trace0)
            horizon = hname.get()
            verify = False
            if (not horizon) or (self.popup.flag==-1):
                return

        if horizon:
            if verify and not tk.messagebox.askyesno("Warning","Remove " + horizon + " horizon?"):
                return
            if isinstance(horizon,str):
                horizon = [horizon]
            if self.get_pickState() and self.horVar.get() in horizon:
                self.set_pickState(False)
            for h in horizon:
                # get index of key to remove annotation
                i = list(self.horizon_paths.keys()).index(h)
                self.ann_list[i].remove()
                del self.ann_list[i]
                # remove path objects, lines, and pick horizons
                self.horizon_lns[h].remove()
                del self.horizon_lns[h]
                del self.horizon_paths[h]
                del self.rdata.pick.horizons[h]
                del self.ln_colors["used"][h]
                if h == self.rdata.pick.srf:
                    self.rdata.pick.set_srf(None)
            # reset horizon and segment variables
            if len(self.horizon_paths) > 0:
                h = list(self.horizon_paths.keys())[-1]
                self.horVar.set(h)
                self.segVar.set(len(self.horizon_paths[h]))
            else:
                self.horVar.set("")
                self.segVar.set(0)
            self.update_pickLabels() 
            self.update_bg()
            self.update_hor_opt_menu()
            self.update_seg_opt_menu()
    

    # init_segment is a method to initialize new pick segment
    def init_segment(self, horizon=None):
        if horizon:
            l = len(self.horizon_paths[horizon])
            self.horizon_paths[horizon][l] = path(np.repeat(np.nan, self.rdata.tnum), np.repeat(np.nan, self.rdata.tnum))
            self.segVar.set(l)
            # update segment options
            self.update_seg_opt_menu()


    # edit selected interpretation segment
    def edit_segment(self, horizon=None, seg=None, verify=True):
        if len(self.horizons) < 1 or self.get_pickState():
            return
        # popup window to get horizon and segment option
        if horizon is None and seg is None:
            if self.popup.flag == 1:
                return
            hname = tk.StringVar()
            hname.set(self.horVar.get())
            s = tk.IntVar()
            s.set(self.segVar.get())
            # create popup window to rename horizon
            popup = self.popup.new(title="Edit Horizon Segment")
            # create horizon dropdown menu
            tk.Label(popup, text="Horizon:").pack(fill="both", expand=True)
            hdropdown = tk.OptionMenu(popup, hname, *[None])
            hdropdown.config(width=20)
            hdropdown.pack(fill="none", expand=True)
            # horizon dropdown with ln colors
            self.color.set(self.ln_colors["str"][self.ln_colors["hex"].index(self.ln_colors["used"][hname.get()])])
            self.set_menu_color(menu=hdropdown, var=hname)
            # trace change in self.color to self.set_menu_color
            trace0 = hname.trace("w", lambda *args, menu=hdropdown, var=hname : self.set_menu_color(menu, var))
            for key, val in self.ln_colors["used"].items():
                hdropdown["menu"].add_command(label=key, foreground=val, activeforeground=val, command=tk._setit(hname, key))

            # create seg dropdown menu
            tk.Label(popup, text="Segment:").pack(fill="both", expand=True)
            sdropdown = tk.OptionMenu(popup, s, *[None])
            sdropdown.config(width=20)
            sdropdown.pack(fill="none", expand=True)
            # trace change in self.color to self.set_menu_color
            trace1 = hname.trace("w", lambda *args, last=True, menu=sdropdown, hvar=hname, svar=s: self.update_seg_opt_menu(last, menu, hvar, svar))
            hname.set(self.horVar.get())
            button = tk.Button(popup, text="OK", command=lambda:self.popup.close(flag=0), width=20).pack(side="left", fill="none", expand=True)
            button = tk.Button(popup, text="Cancel", command=lambda:[hname.set(""), self.popup.close(flag=-1)], width=20).pack(side="left", fill="none", expand=True)
            # wait for window to be closed
            self.parent.wait_window(popup)
            # remove color trace
            hname.trace_vdelete("w", trace0)
            hname.trace_vdelete("w", trace1)
            horizon = hname.get()
            seg = s.get()
            verify = False
            if (not horizon) or (seg is None) or (self.popup.flag==-1):
                return

        # ensure segment belongs to horizon and that interpretations have been made
        if (seg in self.horizon_paths[horizon]):
            if np.isnan(self.horizon_paths[horizon][seg].x).all():
                return
            if verify and not (tk.messagebox.askokcancel("Warning", "Edit interpretation segment " + horizon + "_" + str(seg) + "?", icon="warning")):
                return
            self.edit_flag = True
            self.set_pickState(True)
            # find indices of picked traces
            picks_idx = np.where(~np.isnan(self.horizon_paths[horizon][seg].x))[0]
            # update self.edit_path
            self.edit_path = path(picks_idx, self.horizon_paths[horizon][seg].y[picks_idx])
            # return picked traces to xln list
            self.tmp_horizon_path.x = picks_idx[::20].tolist()
            # return picked samples to yln list
            self.tmp_horizon_path.y = self.horizon_paths[horizon][seg].y[self.tmp_horizon_path.x].tolist()
            # clear saved picks for horizon segment
            self.horizon_paths[horizon][seg].x[picks_idx] = np.nan
            self.horizon_paths[horizon][seg].y[picks_idx] = np.nan
            self.rdata.pick.horizons[horizon][picks_idx] = np.nan
            # reset plotted lines
            self.tmp_horizon_ln.set_data(self.tmp_horizon_path.x, self.tmp_horizon_path.y)
            x,y = utils.merge_paths(self.horizon_paths[horizon])
            self.horizon_lns[horizon].set_data(x,y)
            self.update_bg()


    # remove selected interpretation segment
    def rm_segment(self, horizon=None, seg=None):
        if len(self.horizons) < 1:
            return
        # popup window to get horizon and segment option
        if horizon is None and seg is None:
            if self.popup.flag == 1:
                return
            hname = tk.StringVar()
            hname.set(self.horVar.get())
            s = tk.IntVar()
            s.set(self.segVar.get())
            # create popup window to rename horizon
            popup = self.popup.new(title="Remove Horizon Segment")
            # create horizon dropdown menu
            tk.Label(popup, text="Horizon:").pack(fill="both", expand=True)
            hdropdown = tk.OptionMenu(popup, hname, *[None])
            hdropdown.config(width=20)
            hdropdown.pack(fill="none", expand=True)

            # horizon dropdown with ln colors
            self.color.set(self.ln_colors["str"][self.ln_colors["hex"].index(self.ln_colors["used"][hname.get()])])
            self.set_menu_color(menu=hdropdown, var=hname)
            # trace change in self.color to self.set_menu_color
            trace0 = hname.trace("w", lambda *args, menu=hdropdown, var=hname : self.set_menu_color(menu, var))
            for key, val in self.ln_colors["used"].items():
                hdropdown["menu"].add_command(label=key, foreground=val, activeforeground=val, command=tk._setit(hname, key))
            # create seg dropdown menu
            tk.Label(popup, text="Segment:").pack(fill="both", expand=True)
            sdropdown = tk.OptionMenu(popup, s, *[None])
            sdropdown.config(width=20)
            sdropdown.pack(fill="none", expand=True)
            # trace change in self.color to self.set_menu_color
            trace1 = hname.trace("w", lambda *args, last=True, menu=sdropdown, hvar=hname, svar=s: self.update_seg_opt_menu(last, menu, hvar, svar))
            hname.set(self.horVar.get())
            button = tk.Button(popup, text="OK", command=lambda:self.popup.close(flag=0), width=20).pack(side="left", fill="none", expand=True)
            button = tk.Button(popup, text="Cancel", command=lambda:[hname.set(""), self.popup.close(flag=-1)], width=20).pack(side="left", fill="none", expand=True)
            # wait for window to be closed
            self.parent.wait_window(popup)
            # remove hname traces
            hname.trace_vdelete("w", trace0)
            hname.trace_vdelete("w", trace1)
            horizon = hname.get()
            seg = s.get()
            if (not horizon) or (seg is None) or (self.popup.flag==-1):
                return
        

        # ensure segment belongs to horizon and that interpretations have been made
        if (seg in self.horizon_paths[horizon]):
            if (np.isnan(self.horizon_paths[horizon][seg].x).all()) and not (self.edit_flag):
                return

            if not (tk.messagebox.askokcancel("Warning", "Remove interpretation segment " + horizon + "_" + str(seg) + "?", icon="warning")):
                return
            # if currently editing layer, clear tmp horizon objects
            if self.edit_flag:
                # clear active pick lists
                del self.tmp_horizon_path.x[:]
                del self.tmp_horizon_path.y[:]
                self.tmp_horizon_ln.set_data(self.tmp_horizon_path.x, self.tmp_horizon_path.y)
                self.set_pickState(False)

            # remove horizon interpretation path object for segment
            del self.horizon_paths[horizon][seg]

            # reorder pick segments and path objects if necessary
            l = len(self.horizon_paths[horizon])
            if seg != l:
                for _i in range(seg, l):
                    self.horizon_paths[horizon][_i] = path(self.horizon_paths[horizon][_i + 1].x, self.horizon_paths[horizon][_i + 1].y)
                del self.horizon_paths[horizon][_i + 1]

            self.update_pickLabels() 
            self.update_seg_opt_menu(last=True)
            # reset line object for horizon
            x,y = utils.merge_paths(self.horizon_paths[horizon])
            self.horizon_lns[horizon].set_data(x,y)
            self.update_bg()
    

    # set_pickState is a method to handle the pick state for a given surface as well as pick labels
    def set_pickState(self, state=False):
        horizon = self.horVar.get()
        seg=self.segVar.get()
        # temporary path object length
        l = len(self.tmp_horizon_path.x)                
        # handle pick state set to true
        if state:
            if not horizon:
                self.init_horizon()
                return
            # if pick state was previously true
            if self.get_pickState():
                if l >= 2:
                    self.pick_interp(horizon=horizon,seg=seg)
                    self.plot_picks(horizon=horizon)
                    # if subsequent segment doesn't already exist, initialize
                    if (seg + 1) not in self.horizon_paths[horizon]:
                        self.init_segment(horizon)
                    else:
                        self.segVar.set(seg + 1)
                else:
                    self.clear_last()
            # if pick state was previously false
            else:
                self.startbutton.config(relief="sunken")
                self.stopbutton.config(relief="raised")
                # temporarily disable horizon and segment menus
                self.horMenu.config(state="disabled")
                self.segMenu.config(state="disabled")

        # handle pick state set to false
        else:
            self.startbutton.config(relief="raised")
            self.stopbutton.config(relief="sunken")
            # reactivate horizon and segment menus
            self.horMenu.config(state="active")
            self.segMenu.config(state="active")
            # if pick state was previously true
            if self.get_pickState():
                if l >=  2:
                    # handle previously editing
                    if self.edit_flag:
                        # if no edits made, return to previous segment picks
                        x = self.edit_path.x[::20].tolist()
                        y = self.edit_path.y[::20].tolist()
                        if (x == self.tmp_horizon_path.x) and (y == self.tmp_horizon_path.y):
                            self.tmp_horizon_path.x = self.edit_path.x.tolist()
                            self.tmp_horizon_path.y = self.edit_path.y.tolist()
                    # interpolate and plot picks
                    self.pick_interp(horizon=horizon,seg=seg)
                    self.plot_picks(horizon=horizon)
                    if (seg + 1) not in self.horizon_paths[horizon]:
                        self.init_segment(horizon)
                    else:
                        self.segVar.set(seg + 1)
                # if less than 2 picks, clear
                else:
                    self.clear_last()
                    del self.horizon_paths[horizon][list(self.horizon_paths[horizon].keys())[-1]]
            # reset edit_flag
            self.edit_flag = state
        # update pick_state to current state
        self.pick_state = state
        # set crosshair visibility
        self.set_cross_hair_visible(state)
        # add pick annotations
        self.update_pickLabels()
        self.update_seg_opt_menu()
        self.update_bg()


    # addseg is a method to for user to generate picks
    def addseg(self, event):
        if self.rdata.fpath:
            # store pick trace idx as nearest integer
            pick_trace = int(round(event.xdata))
            # handle picking of last trace - don't want to round up here
            if pick_trace == self.rdata.tnum:
                pick_trace -= 1
            # store pick sample idx as nearest integer
            pick_sample = int(round(event.ydata))
            # ensure pick is within radargram bounds
            if (pick_trace < 0) or (pick_trace > self.rdata.tnum - 1) or \
                (pick_sample < 0) or (pick_sample > self.rdata.snum - 1):
                return

            # pass pick_trace to basemap
            if self.basemap and self.basemap.get_state() == 1:
                self.basemap.plot_idx(self.rdata.fn, pick_trace)

            # print pick info if in debug mode
            if self.debugState == True:
                utils.print_pickInfo(self.rdata, pick_trace, pick_sample, self.eps_r)

            # check if picking state is a go
            if self.get_pickState():
                # if pick_trace falls within other segment for horizon, return
                for path in self.horizon_paths[self.horVar.get()].values():
                    if pick_trace in path.x:
                        return

                # determine if trace already contains pick - if so, replace with current sample
                if pick_trace in self.tmp_horizon_path.x:
                    self.tmp_horizon_path.y[self.tmp_horizon_path.x.index(pick_trace)] = pick_sample

                else:
                    # if pick falls before previous pix, prepend to pick list
                    if (len(self.tmp_horizon_path.x) >= 1) and (pick_trace < self.tmp_horizon_path.x[0]):
                        self.tmp_horizon_path.x.insert(0, pick_trace)
                        self.tmp_horizon_path.y.insert(0, pick_sample)
                        self.update_pickLabels()
                        self.update_bg()

                    # if pick falls in between previous picks, insert in proper location of pick list
                    elif (len(self.tmp_horizon_path.x) >= 1) and (pick_trace > self.tmp_horizon_path.x[0]) and (pick_trace < self.tmp_horizon_path.x[-1]):
                        # find proper index to add new pick 
                        idx = utils.list_insert_idx(self.tmp_horizon_path.x, pick_trace)   
                        self.tmp_horizon_path.x = self.tmp_horizon_path.x[:idx] + [pick_trace] + self.tmp_horizon_path.x[idx:] 
                        self.tmp_horizon_path.y = self.tmp_horizon_path.y[:idx] + [pick_sample] + self.tmp_horizon_path.y[idx:] 

                    # else append new pick to end of pick list
                    else:
                        self.tmp_horizon_path.x.append(pick_trace)
                        self.tmp_horizon_path.y.append(pick_sample)
                        if len(self.tmp_horizon_path.x) == 1:
                            self.update_pickLabels()

                # set picks and draw
                self.tmp_horizon_ln.set_data(self.tmp_horizon_path.x, self.tmp_horizon_path.y)
                self.blit()


    # pick_interp is a method for cubic spline interpolation of twtt between pick locations
    def pick_interp(self, horizon=None, seg=None):
        # get current window size - handle non int entry
        try:
            winSize = self.winSize.get()  
        except:
            winSize = 0
            self.winSize.set(0) 
        # if there are at least two picked points, interpolate
        if len(self.tmp_horizon_path.x) >= 2:
            # cubic spline between picks
            cs = CubicSpline(self.tmp_horizon_path.x, self.tmp_horizon_path.y)
            # generate array between first and last pick indices on current layer
            picked_traces = np.arange(self.tmp_horizon_path.x[0], self.tmp_horizon_path.x[-1] + 1)
            sample = cs(picked_traces).astype(int)
            # if windize >=2, loop over segment and take maximum sample within window of cubic spline interp
            if winSize >= 2:
                for _i in range(len(picked_traces)):
                    sample[_i] = int(sample[_i] - (winSize/2)) + np.argmax(np.abs(self.rdata.dat)[int(sample[_i] - (winSize/2)):int(sample[_i] + (winSize/2)), picked_traces[_i]])
            # add pick interpolation to horizon objects for current segment
            self.horizon_paths[horizon][seg].x[picked_traces] = picked_traces
            self.horizon_paths[horizon][seg].y[picked_traces] = sample
            self.rdata.pick.horizons[horizon][picked_traces] = sample


    # plot_picks is a method to remove current pick list and add saved picks to plot
    def plot_picks(self, horizon=None):
        # remove temporary picks
        del self.tmp_horizon_path.x[:]
        del self.tmp_horizon_path.y[:]
        self.tmp_horizon_ln.set_data(self.tmp_horizon_path.x, self.tmp_horizon_path.y)
        x,y = utils.merge_paths(self.horizon_paths[horizon])
        self.horizon_lns[horizon].set_data(x,y)


    # clear last pick
    def clear_last(self):
        if self.pick_state == True:
            if  len(self.tmp_horizon_path.x) >= 1: 
                del self.tmp_horizon_path.x[-1:]
                del self.tmp_horizon_path.y[-1:]
                # reset self.tmp_horizon_ln, then blit
                self.tmp_horizon_ln.set_data(self.tmp_horizon_path.x, self.tmp_horizon_path.y)
                self.blit()


    # set_picks is a method to set imported pick arrays to horizon_paths
    def set_picks(self, horizon=None):
        if horizon:
            # set horizon color to first unused color
            color = [c for c in self.ln_colors["hex"] if c not in list(self.ln_colors["used"].values())][0]
            self.color.set(self.ln_colors["str"][self.ln_colors["hex"].index(color)])
            self.init_horizon(horizon=horizon,skip_array=True)
            # split horizon array into segments
            idx = utils.nonan_idx_array(self.rdata.pick.horizons[horizon])
            clumps = utils.clump_array(idx)
            for _i, clump in enumerate(clumps):
                clump_arr = np.asarray(clump).astype(int)
                self.horizon_paths[horizon][_i].x[clump_arr] = clump_arr
                self.horizon_paths[horizon][_i].y[clump_arr] = self.rdata.pick.horizons[horizon][clump_arr]
                self.init_segment(horizon=horizon)
            x,y = utils.merge_paths(self.horizon_paths[horizon])
            self.horizon_lns[horizon].set_data(x,y)


    # update the horizon menu
    def update_hor_opt_menu(self):
        self.horizons = list(self.horizon_paths.keys())
        self.horMenu["menu"].delete(0, "end")
        for i, horizon in enumerate(self.horizons):
            c = self.horizon_lns[horizon].get_color()
            self.horMenu["menu"].add_command(label=horizon, foreground=c, activeforeground=c, command=tk._setit(self.horVar, horizon))
            # add horizon colors to ln_colors["used"] dict for use in wvpick
            self.ln_colors["used"][horizon] = c


    # update the horizon segment menu based on how many segments exist for given segment
    def update_seg_opt_menu(self, last=False, menu=None, hvar=None, svar=None, *args):
        if menu is None:
            menu = self.segMenu
        if hvar is None:
            hvar = self.horVar
        if svar is None:
            svar = self.segVar
        horizon = hvar.get()
        menu["menu"].delete(0, "end")
        if horizon:
            for seg in sorted(self.horizon_paths[horizon].keys()):
                menu["menu"].add_command(label=seg, command=tk._setit(svar, seg))
            # set segment selection to last
            if last:
                svar.set(seg)


    # update_pickLabels is a method to create annotations for picks
    def update_pickLabels(self):
        for _i in self.ann_list:
            _i.remove()
        del self.ann_list[:]
        # get x and y locations for placing annotation
        # first check tmp_horizon paths
        if len(self.tmp_horizon_path.x) > 0:
            x = self.tmp_horizon_path.x[0]
            y = self.tmp_horizon_path.y[0]
            ann = self.ax.text(x-25,y+75, self.horVar.get() + "_" + str(self.segVar.get()), bbox=dict(facecolor='white', alpha=0.5), horizontalalignment='right', verticalalignment='top')
            self.ann_list.append(ann)
            if not self.ann_vis:
                ann.set_visible(False)
        for horizon, item in self.horizon_paths.items():
            for seg, path in item.items():
                x = np.where(~np.isnan(path.x))[0]
                if x.size == 0:
                    continue
                x = x[0]
                y = path.y[x]
                ann = self.ax.text(x-25,y+75, horizon + "_" + str(seg), bbox=dict(facecolor='white', alpha=0.5), horizontalalignment='right', verticalalignment='top')
                self.ann_list.append(ann)
                if not self.ann_vis:
                    ann.set_visible(False)


    def show_labels(self, vis=None):
        if vis is not None:
            self.ann_vis = vis
        if self.ann_vis:
            for _i in self.ann_list:
                _i.set_visible(True)
        else:
            for _i in self.ann_list:
                _i.set_visible(False)
        self.update_bg()


    # show_picks is a method to toggle the visibility of picks on
    def show_picks(self, vis=None):
        if vis is not None:
            self.pick_vis = vis
        self.show_artists(self.pick_vis)
        self.safe_draw()
        self.fig.canvas.blit(self.ax.bbox)


    # show plotted lines - don't override cursor crosshair visibility
    def show_artists(self,val=True,force=False):
        for _i in self.ax.lines:
            if force:
                _i.set_visible(val)
            else:
                if _i == self.horizontal_line:
                    pass
                elif _i ==self.vertical_line:
                    pass
                else:
                    _i.set_visible(val)


    # temporarily disconnect the draw_event callback to avoid recursion
    def safe_draw(self):
        canvas = self.fig.canvas
        canvas.mpl_disconnect(self.draw_cid)
        canvas.draw()
        self.draw_cid = canvas.mpl_connect("draw_event", self.update_bg)


    # when the figure is resized, hide picks, draw everything, and update the background image
    def update_bg(self, event=None):
        # temporarily hide artists
        self.show_artists(False, force=True)
        self.safe_draw()
        self.axbg = self.dataCanvas.copy_from_bbox(self.ax.bbox)
        # return artists visiblity to former state
        self.show_artists(self.pick_vis,self.get_pickState())
        self.blit()


    # update the figure, without needing to redraw the "axbg" artists
    def blit(self):
        self.fig.canvas.restore_region(self.axbg)
        for _i in self.ax.lines:
            self.ax.draw_artist(_i)
        for _i in self.ann_list:
            self.ax.draw_artist(_i)
        self.fig.canvas.blit(self.ax.bbox)


    # clear_canvas is a method to clear the data canvas and figures to reset app
    def clear_canvas(self):
        # clearing individual axis objects seems to keep a history of these objects and causes axis limit issues when opening new track
        self.ax.cla()


    # get_basemap is a method to hold the basemap object passed from gui
    def get_basemap(self, basemap):
        self.basemap = basemap


    # onpress gets the time of the button_press_event
    def onpress(self,event):
        if event.inaxes == self.ax:
            self.dblclick = event.dblclick
            self.time_onclick = time.time()


    # onrelease calls addseg() if the time between the button press and release events
    # is below a threshold so that segments are not drawn while trying to zoom or pan
    def onrelease(self,event):
        if event.inaxes == self.ax:
            if event.button == 1 and ((time.time() - self.time_onclick) < 0.25):
                self.addseg(event)
                # if double clicked and currently picking, end current segment
                if self.dblclick and self.get_pickState():
                    self.set_pickState(False)
            self.time_onclick = time.time()


    # set_cross_hair_visible
    def set_cross_hair_visible(self, visible):
        self.horizontal_line.set_visible(visible)
        self.vertical_line.set_visible(visible)


    # on_mouse_move blit crosshairs
    def on_mouse_move(self, event):
        if self.rdata:
            x, y = event.xdata, event.ydata
            self.horizontal_line.set_ydata(y)
            self.vertical_line.set_xdata(x)
            self.ax.figure.canvas.restore_region(self.axbg)
            self.ax.draw_artist(self.horizontal_line)
            self.ax.draw_artist(self.vertical_line)
            self.blit()


    # update_figsettings
    def update_figsettings(self, figsettings=None):
        if figsettings:
            self.figsettings = figsettings
        # update all label sizes
        fs = self.figsettings["fontsize"].get()
        plt.rcParams.update({'font.size': fs})
        for item in ([self.ax.xaxis.label, self.ax.yaxis.label, self.secaxx.xaxis.label, self.secaxy0.yaxis.label, self.secaxy1.yaxis.label] +
                    self.ax.get_xticklabels() + self.ax.get_yticklabels() + self.secaxx.get_xticklabels() + self.secaxy0.get_yticklabels() + self.secaxy1.get_yticklabels() + self.ann_list):
            item.set_fontsize(fs)
        self.ax.title.set_size(fs)
        self.s_cmin.label.set_fontsize(fs)
        self.s_cmax.label.set_fontsize(fs)
        self.s_cmin.valtext.set_fontsize(fs)
        self.s_cmax.valtext.set_fontsize(fs)
        self.cmap_reset_button.label.set_fontsize(fs)
        # update title visibility
        self.ax.title.set_visible(self.figsettings["figtitle"].get())
        # update x-axes
        val = self.figsettings["figxaxis"].get()
        self.ax.xaxis.set_visible(val)
        self.secaxx.axis(self.figsettings["figxaxis"].get())
        # update y-axes
        val = self.figsettings["figyaxis"].get()
        self.ax.yaxis.set_visible(val)
        self.secaxy0.axis(val)
        self.secaxy1.axis(val)
        self.set_cmap(self.figsettings["cmap"].get())
        self.fig.canvas.draw()


    # get debug state from gui settings
    def set_debugState(self, debugState):
        self.debugState = debugState


    # get eps_r setting from gui settings
    def set_eps_r(self, eps_r):
        self.eps_r = eps_r


    # export_fig is a method to receive the pick save location from gui export the radar figure
    def export_fig(self, f_saveName):
        # zoom out to full rgram extent to save pick image
        self.fullExtent()
        self.verticalClip(self.figsettings["figclip"][0].get(), self.figsettings["figclip"][1].get())
        w0 ,h0 = self.fig.get_size_inches()    # get pre-save figure size
        w, h = self.figsettings["figsize"].get().split(",")
        self.fig.set_size_inches((float(w),float(h)))    # set figsize to wide aspect ratio
        # temporarily turn sliders to invisible for saving image
        self.ax_cmax.set_visible(False)
        self.ax_cmin.set_visible(False)
        self.reset_ax.set_visible(False)
        self.secaxy1.set_visible(False)
        # hide pick annotations
        vis = self.ann_vis
        if vis:
            self.show_labels(vis=False)  

        # save data fig with and without picks
        if self.im_status.get() ==1:
            self.show_data()
        self.show_picks(vis=False)
        export.fig(f_saveName.rstrip(f_saveName.split("/")[-1]) + self.rdata.fn + ".png", self.fig)
        self.show_picks(vis=True)
        export.fig(f_saveName, self.fig)
    
        # save sim fig if sim exists
        if self.rdata.flags.sim:
            self.show_sim()
            # ensure picks are hidden
            # self.pick_vis.set(False)
            self.show_picks(vis=False)
            export.fig(f_saveName.rstrip(f_saveName.split("/")[-1]) + self.rdata.fn + "_sim.png", self.fig)
            # self.pick_vis.set(True)
            self.show_picks(vis=True)
            self.show_data()

        # return figsize to intial values and make sliders visible again
        self.fig.set_size_inches((w0, h0))
        self.fullExtent()
        self.ax_cmax.set_visible(True)
        self.ax_cmin.set_visible(True)
        self.reset_ax.set_visible(True)
        self.secaxy1.set_visible(True)
        if vis:
            self.show_labels(True)   
        self.fig.canvas.draw()


class path():
    # initialize a path object to hold x,y paths for plotting - list or array like
    def __init__(self, x=None, y=None):
        self.x = x      # x array/list
        self.y = y      # y array/list


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license

class raguError(Exception):
    """Used for various ragu exceptions."""


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
radar data ingest wrapper
"""
### imports ###
from ragu.ingest import ingest_oibAK, ingest_groundhog, ingest_uaf_kentech, ingest_pulseekko, ingest_gssi, ingest_sharad, ingest_marsis, ingest_marsis_ipc, ingest_lrs, ingest_cresis_rds, ingest_cresis_snow, ingest_rimfax
from ragu.tools import utils
import numpy as np
import pandas as pd
import tkinter as tk
import fnmatch

class ingest:
    # ingest is a class which builds a dictionary holding data and metadata from the file
    def __init__(self, fpath):
        # ftype is a string specifying filetype
        # valid options -
        # hdf5, mat, segy, img, dat, dt1, dzt, lbl
        valid_types = ["h5", "mat", "img", "dat", "dt1", "dzt", "gpz", "lbl" ,"csv"]
        ftype = fpath.split(".")[-1].lower()
        
        if (ftype not in valid_types):

            raise ValueError("Invalid file type specifier: " + 
                ftype + "\nValid file types: " + str(valid_types))

        self.fpath = fpath
        self.ftype = ftype


    def read(self, simpath=None, navcrs='+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs', body='earth'):
        # wrapper method for reading in a file
        # better ways to do this than an if/else
        # but for a few file types this is easier
        if (self.ftype == "h5"):
            try:
                self.rdata = ingest_oibAK.read_h5(self.fpath, navcrs, body)
            except:
                try:
                    self.rdata = ingest_groundhog.read_h5(self.fpath, navcrs, body)
                except:
                    self.rdata = ingest_uaf_kentech.read_h5(self.fpath, navcrs, body)
        elif (self.ftype == "mat"):
            try:
                self.rdata = ingest_cresis_snow.read_mat(self.fpath, navcrs, body)
            except:
                try:
                    self.rdata = ingest_cresis_rds.read_mat(self.fpath, navcrs, body)
                except:
                    self.rdata = ingest_oibAK.read_mat(self.fpath, navcrs, body)
        elif (self.ftype == "img"):
            try:
                self.rdata = ingest_sharad.read(self.fpath, simpath, navcrs, body)
            except:
                try:
                    self.rdata = ingest_lrs.read(self.fpath, simpath, navcrs, body)
                except:
                    try:
                        self.rdata = ingest_marsis.read(self.fpath, simpath, navcrs, body)
                    except:
                        self.rdata = ingest_marsis_ipc.read(self.fpath, simpath, navcrs, body)
        elif (self.ftype == "dat"):
            self.rdata = ingest_marsis.read(self.fpath, simpath, navcrs, body)
        elif (self.ftype == "csv"):
            self.rdata = ingest_rimfax.read(self.fpath, navcrs, body)
        elif (self.ftype == "dt1"):
            self.rdata = ingest_pulseekko.read_dt1(self.fpath, navcrs, body)
        elif (self.ftype == "gpz"):
            raise ValueError("Error: \tPulseEKKO GPZ project file ingester currently in development.\n\tExport lineset from EKKO_Project to read DT1 files with RAGU")
            # self.rdata = ingest_pulseekko.partition_project_file(self.fpath, navcrs, body)
        elif (self.ftype == "dzt"):
            self.rdata = ingest_gssi.read(self.fpath, navcrs, body)

        else:
            print("File reader for format {} not built yet".format(self.ftype))
            exit(1)

        print("----------------------------------------")
        print("Loaded: " + self.rdata.fn)

        # add ingest commands to log
        self.rdata.log('igst = ingest.ingest("{}")'.format(self.fpath))
        self.rdata.log('rdata = igst.read("{}","{}","{}")'.format(simpath,navcrs,body))

        return self.rdata


    # import_pick is a method of the ingester class which loads in existing picks from a text file
    def import_pick(self, fpath, uid, force=False):
        if fpath.endswith("csv"):
            dat = pd.read_csv(fpath, comment='#')
            if dat.shape[0] != self.rdata.tnum:
                raise ValueError("import_pick error:\t pick file size does not match radar data")
                return
            horizons = []
            keys = fnmatch.filter(dat.keys(), "*sample*")
            if len(keys) >= 1:
                for horizon in keys:
                    # handle merged and individual pick files
                    # merged files will have ["horizon_sample"] keys, while individual will have "horizon" in pick file name and ["sample"] in keys
                    horizon = horizon.split("_")
                    if len(horizon) > 1:
                        horizon = horizon[0]
                        sample = dat[horizon + "_sample"].to_numpy() 
                    
                    else:
                        # get horizon name from pick file name
                        fnl = len(self.rdata.fn) + 1
                        extl = len(uid) + 4
                        horizon = fpath.split("/")[-1][fnl:-extl]
                        sample = dat["sample"].to_numpy() - self.rdata.flags.sampzero
                    
                        # account for any already applied tzero
                    sample -= self.rdata.flags.sampzero

                    if (not force) and (not tk.messagebox.askyesno("Import Horizon","Import " + str(horizon) + " horizon?")):
                        continue

                    if horizon in self.rdata.pick.horizons.keys():
                        # if imported horizon same as existing, continue
                        if utils.nan_array_equal(self.rdata.pick.horizons[horizon], sample):
                            continue
                        # elif imported horizon different form existing, and existing is not all nan, append '_imported' on name
                        elif not np.isnan(self.rdata.pick.horizons[horizon]).all():
                            horizon = horizon + "_imported"
                    self.rdata.pick.horizons[horizon] = sample
                    horizons.append(horizon)

        return  horizons


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_template is a RAGU data ingest template. Follow this basic template format and modify to read your radar data type.

NOTE: ingest/__init__.py must also be modified after creating a new ingester for RAGU to be able to ingest your data type 
ALSO: nav/navparse.py will need an additional method for reading your navigation data
"""
### necessary imports, different data types may require additional ###
from ragu.radar import garlic
from ragu.nav import navparse
from ragu.tools import utils
import h5py, fnmatch
import numpy as np
import scipy as sp
import sys

# method to ingest your data
def read_dat(fpath, navcrs, body):
    # initialize the radar data object - garlic() takes the data file path
    rdata = garlic(fpath)

    # define the rdar data file name - str
    rdata.fn = fpath.split("/")[-1][:-3]

    # name the data type - str
    rdata.dtype = "blah"

    # read in .h5 file
    f = h5py.File(rdata.fpath, "r")                      

    ############################################################################
    #### necessary radar attributes - modify these based on your data type   ###
    ############################################################################
    rdata.snum = None       # samples per trace in rgram, int
    rdata.tnum = None       # number of traces in rgram, int
    rdata.fs =  None        # sampling frequency (Hz), float
    rdata.prf = None        # pulse repition frequency (Hz), float
    data = None             # 2D radar data array of shape ((rdata.snum x rdata.tnu,)) - pulse compressed if chirped data- this should be amplitude units, float

    # parse nav - you may need to create a navparse method for reading your nav data (replace "getnav" with the name of your nav parser method)
    rdata.geocrs = navcrs
    rdata.xyzcrs = navparse.xyzsys[body]
    rdata.navdf = navparse.getnav(fpath, navcrs, body)
    ############################################################################


    ############################################################################
    ############################# optional changes #############################
    ############################################################################
    rdata.nchan = 1     # number of data channels - don't need to touch this unless multichannel data, int
    sim = None              # 2D radar clutter simulation of shape ((rdata.snum x rdata.tnu,)) - this should be amplitude units, float
    # assign signal info - add more if you'd like. These show up on a banner at the bottom of the RAGU GUI window
    rdata.info["Signal Type"] = 'Impulse'
    rdata.info["CF [MHz]"] =  5 * 1e-6
    rdata.info["Pulse Length [\u03BCs]"] = 2.5 * 1e6
    rdata.info["Sampling Frequency [MHz]"] = rdata.fs * 1e-6
    rdata.info["PRF [kHz]"] = rdata.prf * 1e-3
    ############################################################################




    # shouldn't need to touch these after changing the above
    rdata.dt = 1/rdata.fs               # sampling interval, sec
    rdata.set_dat(data)                 # assign amplitude data array as input radar data
    rdata.set_proc(rdata.get_dat())     # this stores a processed data array - data is dB'd for display purposes
    if sim:
        rdata.set_sim(sim)              # clutter sim array

    rdata.set_twtt()                    # get twtt for fast-time axis
    rdata.check_attrs()                 # check radar data attributes to make sure everything is garlicy

    return rdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_uaf_kentech is a module developed to ingest Martin Truffer's Kentech radar sounding data. 
"""
### imports ###
from ragu.radar import garlic
from ragu.nav import navparse
from ragu.tools import utils
import h5py, fnmatch
import numpy as np
import scipy as sp
import sys

# method to ingest OIB-AK radar hdf5 data format
def read_h5(fpath, navcrs, body):
    rdata = garlic(fpath)
    rdata.fn = fpath.split("/")[-1][:-3]
    rdata.dtype = "uaf_kentech"

    # read in .h5 file
    f = h5py.File(rdata.fpath, "r")                      

    # h5 radar data group structure        
    # |-raw
    # |  |-rx0
    # |  |-tx0    
    # |  |-loc0
    # |-ext
    # |  |-nav0
    # |  |-srf0
    # |-drv
    # |  |-proc0
    # |  |-clutter0
    # |  |-pick

    # pull necessary raw group data
    rdata.snum = int(f["raw"]["rx0"].attrs["samplesPerTrace"])                  # samples per trace in rgram
    rdata.tnum = int(f["raw"]["rx0"].attrs["numTrace"])                         # number of traces in rgram 
    rdata.fs = f["raw"]["rx0"].attrs["samplingFrequency"]                       # sampling frequency
    rdata.dt = 1/rdata.fs                                                       # sampling interval, sec
    rdata.prf = 1000                                                            # pulse repition frequency, Hz
    rdata.nchan = 1

    # pull radar proc and sim arrayss
    rdata.set_dat(f["drv/proc0"][:])                                            # pulse compressed array
    rdata.set_proc(np.abs(rdata.get_dat()))

    rdata.set_twtt()
    # assign signal info
    # rdata.info["Signal Type"] = f["raw"]["tx0"].attrs["signal"].capitalize() 
    # rdata.info["CF [MHz]"] = f["raw"]["tx0"].attrs["centerFrequency"][0] * 1e-6
    # if rdata.info["Signal Type"] == "Chirp":
    #     rdata.info["Badwidth [%]"] = f["raw"]["tx0"].attrs["bandwidth"][0] * 100
    #     rdata.info["Pulse Length [\u03BCs]"] = f["raw"]["tx0"].attrs["length"][0] * 1e6
    # rdata.info["Sampling Frequency [MHz]"] = rdata.fs * 1e-6
    # rdata.info["PRF [kHz]"] = rdata.prf * 1e-3

    # parse nav
    rdata.navdf = navparse.getnav_uaf_kentech(fpath, navcrs, body)

    f.close()                                                   # close the file

    rdata.check_attrs()

    return rdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_oibAK is a module developed to ingest NASA OIB-AK radar sounding data. 
primary data format is hdf5, however some older data is still being converted over from .mat format
"""
### imports ###
from ragu.radar import garlic
from ragu.nav import navparse
from ragu.tools import utils
import h5py, fnmatch
import numpy as np
import scipy as sp
import sys

# method to ingest groundhog hdf5 data format
def read_h5(fpath, navcrs, body):
    rdata = garlic(fpath)
    rdata.fn = fpath.split("/")[-1][:-3]

    # read in .h5 file
    f = h5py.File(rdata.fpath, "r")                      

    # get appropriate data grou
    grps = f.keys()
    if "proc" in grps:
        grp = "proc"
    elif "restack" in grps:
        grp = "restack"
    else: 
        grp = "raw"


    # pull necessary raw group data
    rdata.fs = f[grp]["rx0"].attrs["fs"]                             # sampling frequency
    rdata.dt = 1/rdata.fs
    try:                                                            # sampling interval, sec
        rdata.prf = f[grp]["rx0"].attrs["prf"]                       # pulse repition frequency, Hz
    except KeyError:
        rdata.prf = 0

    rdata.nchan = 1


    rdata.set_dat(f[grp]["rx0"][:].astype(float))

    # get system info
    try:
        if "Blue Systems" in f[grp]["rx0"].attrs["system"]:
            rdata.dtype = "bsi"
            try:
                rdata.set_sim(f["drv"]["clutter0"][:])              # simulated clutter array
            except KeyError:
                pass
    except KeyError:
        rdata.dtype = "ghog"

    rdata.dbit = False

    rdata.set_proc(rdata.get_dat())
    rdata.snum, rdata.tnum = rdata.get_dat().shape

    rdata.set_twtt()

    # parse nav
    rdata.geocrs = navcrs
    rdata.xyzcrs = navparse.xyzsys[body]
    rdata.navdf = navparse.getnav_groundhog(fpath, navcrs, body)

    # if rx and tx each have gps, we'll store antenna sep per trace
    if "asep" in rdata.navdf:
        rdata.asep = rdata.navdf["asep"].to_numpy()

    # groudnhog rx triggers on arrival of airwave - get number of traces pre-trigger to vertically shift the data accordingly
    try:
        pt = f[grp]["rx0"].attrs["pre_trig"]
    except KeyError:
        pt = f[grp]["rx0"].attrs["pre_trigger"]

    rdata.flags.sampzero = pt
    if pt>0:
        rdata.tzero_shift()

    # pull some info
    rdata.info["Signal Type"] = "Impulse" 
    rdata.info["Sampling Frequency [MHz]"] = rdata.fs * 1e-6
    rdata.info["PRF [kHz]"] = rdata.prf * 1e-3
    rdata.info["Stack"] = f[grp]["rx0"].attrs["stack"]
    rdata.info["Antenna Separation [m]"] = round(np.nanmean(rdata.asep))

    rdata.check_attrs()                                         # check basic rdata attributes
    f.close()                                                   # close file


    # initialize surface arrays - if groundhog, surface is at zero after shifting by pretrigger amount. If BSI airIPR initialize nan array for surface pick horizon and elevation array
    if rdata.dtype == "ghog":
        # define surface horizon name to set index to zeros
        arr = np.zeros(rdata.tnum)
        # srf_elev is the same as GPS recorded elev
        rdata.set_srfElev(dat = rdata.navdf["elev"].to_numpy())
    elif rdata.dtype == "bsi":
        arr = np.repeat(np.nan, rdata.tnum)
        rdata.set_srfElev(dat = arr)

        # optional processing defaults
        # rdata.filter(btype='bandpass', lowcut=2.5e6, highcut=20e6, order=5, direction=0)
        # rdata.removeSlidingMeanFFT(window=250)
        rdata.tpowGain(1.5)

    # init surface and bed horizons
    rdata.pick.horizons["srf"] = arr
    rdata.pick.set_srf("srf")
    arr = np.repeat(np.nan, rdata.tnum)
    rdata.pick.horizons["bed"] = arr

    return rdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
utility functions for RAGU
"""
### imports ###
from ragu.tools.constants import *
import numpy as np
import pandas as pd
import geopandas as gpd
import tkinter as tk
import sys, h5py, fnmatch, copy

# get_srf is a function for auto-detecting a radargram surface horizon
def get_srf(dat_array, sig_type=None):
    if sig_type == "Chirp":
        # take max power to define surface 
        max_idx = np.nanargmax(dat_array[10:,:], axis = 0) + 10
        # remove outliers
        not_outlier = remove_outliers(max_idx)
        # interpolate, ignoring outliers
        x = np.arange(dat_array.shape[1])
        return np.interp(x, x[not_outlier], max_idx[not_outlier])
    elif sig_type == "Impulse":
        C = np.zeros_like(dat_array)	# create empty criteria array to localize surface echo for each trace
        gradient = np.gradient(dat_array, axis = 0) # find gradient of each trace in RGRAM

        # criteria for surface echo - indicator is Pt * dPt-1/dt, 
        # where P is the signal energy applied on each grame sample (t)
        # indicator weights energy of a sample by the derivative preceding it

        # vectorized criteria calculation
        C[1:,:] = dat_array[1:,:]*gradient[:-1,:]
        C_max = np.argmax(C, axis = 0)	# find indices of max critera seletor for each column
        return C_max

# remove_outliers is a function to remove outliers from an array
# returns bool array
def remove_outliers(array):
    mean = np.mean(array)
    standard_deviation = np.std(array)
    distance_from_mean = abs(array - mean)
    max_deviations = 2
    not_outlier = distance_from_mean < max_deviations * standard_deviation
    return not_outlier


# delete_savedPicks is a method to clear saved picks from an hdf5 data file
def delete_savedPicks(fpath):
    if fpath.endswith("h5"):
        f =  h5py.File(fpath, "a")
        if "twtt_bed" in f["drv"]["pick"].keys():
            del f["drv/pick"]["twtt_bed" ]
        f.close()


# list_insert is a function to return the element at which to insert a new item to a sorted list
def list_insert_idx(list, n): 
    # search for the position 
    for i in range(len(list)): 
        if list[i] > n: 
            index = i 
            break
    return index


# find nearest value in array to arg
def find_nearest(a, val):
    idx = (np.abs(a - val)).argmin()
    return idx


# sort dictionary full of numpy arrays using array mean value
def sort_array_dict(a, srf=None):
    if len(a) > 1:
        b = copy.deepcopy(a)
        out = {}
        sort_dict = {}
        # if surface horizon, place first
        if srf:
            out[srf] = b[srf]
            b.pop(srf, None)

        # stack all arrays in dictionary
        stacked = np.vstack((list(b.values())))
        # see if an index is occupied with a value for all arrays
        # if so, just sort all arrays based on value for that index
        non_nan_cols_array = np.all(~np.isnan(stacked),axis=0)
        if np.any(non_nan_cols_array):
            idx = np.where(non_nan_cols_array)[0][0]
            for key, arr in b.items():
                sort_dict[key] = arr[idx]

        else:
            # sort array elements by row
            row_sorted_arrays = np.argsort(stacked,axis=0).astype(np.float64)
            # replace nan values
            row_sorted_arrays[np.isnan(stacked)] = np.nan
            # add up sorting to get sum and sort all arrays by the sum
            total = np.nansum(row_sorted_arrays,axis=1)
            for i, (key, arr) in enumerate(b.items()):
                sort_dict[key] = total[i]

        keys = [k for k, v in sorted(sort_dict.items(), key=lambda item: item[1])]
        for key in keys:
            out[key] = a[key]

        return out

    else:
        return a


# clump_array removing nan chunks
def clump_array(a):
    return [a[s] for s in np.ma.clump_unmasked(np.ma.masked_invalid(a))]


# nonan_idx_array returns an array with the index of non-nan values
def nonan_idx_array(a):
    out = np.repeat(np.nan, a.shape[0])
    mask = np.where(~np.isnan(a))[0]
    out[mask] = mask
    return out


# nan_array_equal is a method to determine if two arrays which may contain nan values are equivalent
def nan_array_equal(a, b):
    try:
        return ((a == b) | (np.isnan(a) & np.isnan(b))).all()
    except ValueError:
        return


def nan_array_sum(a, axis=0):
    """
    sum across an array with nans, treating the nan's as zero unless all elements across specified axis are nan
    INPUT:
    a           data array
    axis        axis to sum across

    OUTPUT:
    out         summed output array
    """
    out = np.nansum(a,axis=axis)
    out[np.all(np.isnan(a), axis=axis)] = np.nan    
    return out


def merge_paths(a):
    """
    merge picks from all segment paths in horizon dicitonary
    INPUT:
    a           dictionary containing impick path objects for each key 

    OUTPUT:
    (x,y)       tuple of x,y merged path data
    """
    l = list(a.values())
    x = nan_array_sum(np.stack([path.x for path in l]))
    y = nan_array_sum(np.stack([path.y for path in l]))
    return (x,y)


# compare_horizon_paths between two dictionaries
def compare_horizon_paths(dicta=None, dictb=None):
    out = True
    if len(dicta) == 0 & len(dictb) == 0:
        return True
    for horizon in dicta.keys():
        for seg in dicta[horizon].keys():
            if nan_array_equal(dicta[horizon][seg].y, dictb[horizon][seg].y):
                continue
            else:
                out = False
                break
    return out


# srfpick2elev function
def srfpick2elev(samp, twtt_wind, elev, dt, tnum, asep):
    """
    calculate surface elevation from updated surface pick, subtracting distance form navdf["elev"] to surface
    #: np.ndarray(tnum,) samp, surface index per trace [samle #]
    #: np.ndarray(tnum,) twtt_wind, twtt_window between radar platform and sample 0 in radargram
    #: np.ndarray(tnum,) elev, radar elevation per trace [m.a.s.l.]
    #: float dt, spacing between samples in travel time [seconds]
    #: int tnum, the number of traces in the file
    #: np.ndarray(tnum) or single floating point number asep, antenna separation in meters
    """
    # check shapes
    if samp.shape == twtt_wind.shape == elev.shape == (tnum,):
        twtt = sample2twtt(samp, dt) + twtt_wind

        # if a time-zero shift was applied, dist should equal zero for ground based radar systs
        if (twtt==0.).all():
            dist = 0
        else:
            dist = twtt2depth(twtt, asep, eps_r=1)
        srfElev = elev - dist

        return srfElev


# twtt2depth function
def twtt2depth(twtt, asep, eps_r=3.15):
    v = C/np.sqrt(eps_r)
    depth = np.sqrt((v / 2 * (twtt + (asep / C)))**2 - (asep / 2)**2)       # convert travel time to depth, accounting for any antenna separation - if asetp==0 then simplifies to depth = twtt*v/2 (EQ 1 from Armstrong et al., 2022)
    return depth


# depth2twtt function
def depth2twtt(depth, asep, eps_r=3.15):
    v = C/np.sqrt(eps_r)
    twtt = 2 / v * (np.sqrt((depth**2) + (asep / 2)**2)) - (asep / C)       # if asetp==0 then simplifies to twtt = depth*2/v ( rearangment of EQ 1 from Armstrong et al., 2022)
    return twtt


# twtt2sample
def twtt2sample(array, dt):
    sample_array = np.rint(array / dt)
    return sample_array


# sample2twtt
def sample2twtt(array, dt):
    twtt_array = array * dt
    return twtt_array


# amp2powdB
def amp2powdB(array):
    tmp = np.power(array,2)
    # mask zero-power values
    tmp[tmp == 0] = np.nan
    out = 10*np.log10(tmp)
    return out


# powdB2amp
def powdB2amp(array):
    return np.power(10, (array / 20))


# pkampwind
def pkampwind(array, idx, windsize):
    # loop through array columns and find peak greater than 1  row idx within window of given idx
    # use gradient to find maximum index where absolute value of derivative is greater than 1 sigma
    out = np.repeat(np.nan, array.shape[1])
    grad = np.gradient(array, axis = 1)
    std = np.std(array, axis = 0)
    for _i in range(array.shape[1]):
        if np.isnan(idx[_i]):
            continue
        out[_i] = int(idx[_i] - (windsize/2)) + np.argmax(np.abs(grad[int(idx[_i] - (windsize/2)):int(idx[_i] + (windsize/2)), _i]) > std[_i])
    return out


def print_pickInfo(data, trace, sample, eps_r=3.15):
    v = C/(np.sqrt(eps_r))        # EM wave veloity in ice - for thickness calculation

    if data.pick.srf:
        srfTwtt = sample2twtt(data.pick.horizons[data.pick.srf][trace],data.dt)
        elev = data.navdf["elev"].iloc[trace]    
        srfElev = data.srfElev[trace]
        subsrfTwtt = sample2twtt(sample, data.dt)
        thick = (((subsrfTwtt - srfTwtt) * v) / 2)
        subsrfElev = srfElev - thick

        # create table header
        header = "{:<12s} {:<12s} {:<12s} {:<12s} {:<12s} {:<12s} {:<12s} {:<12s}".format\
            ("trace", "sample", "radar elev", data.pick.srf + " elev", "agl" + data.pick.srf + " TWTT", "pick TWTT", "pick elev", "thick")
        markers = "-" * len(header)
        header = [markers, header, markers]
        dat = ["{:<12d} {:<12d} {:<12.4f} {:<12.4f} {:<12.4f} {:<12.4e} {:<12.4e} {:<12.4f} {:<12.4f}".format(trace, sample, elev, srfElev, elev-srfElev, srfTwtt, subsrfTwtt, subsrfElev, thick)]
        print("\n".join(header + dat)+"\n")

    else:
        subsrfTwtt = sample2twtt(sample, data.dt)

        # create table header
        header = "{:<12s} {:<12s} {:<12s}".format\
            ("trace", "sample", "pick TWTT")
        markers = "-" * len(header)
        header = [markers, header, markers]
        dat = ["{:<12d} {:<12d} {:<12.4e} ".format(trace, sample, subsrfTwtt)]
        print("\n".join(header + dat)+"\n")


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_cresis_rds is a module developed to ingest CReSIS snow accumulation radar data. 
"""
### imports ###
from ragu.radar import garlic
from ragu.nav import navparse
from ragu.tools import utils
import h5py, fnmatch
import numpy as np
import sys
import matplotlib.pyplot as plt
# method to ingest CReSIS snow radar data
def read_mat(fpath, navcrs, body):
    rdata = garlic(fpath)
    rdata.fn = fpath.split("/")[-1][:-4]
    rdata.dtype = "cresis_snow"
    f = h5py.File(rdata.fpath, "r")                      

    # assign signal info
    rdata.info["System"] = str(f["param_records"]["radar_name"][:], 'utf-16')
    if "snow" not in rdata.info["System"]:
        raise ValueError("Not snow radar data")
        return

    rdata.set_dat(np.array(f["Data"][:].T))
    rdata.set_proc(np.abs(rdata.get_dat()))
    rdata.snum = rdata.dat.shape[0]                                                 # samples per trace in rgram
    rdata.tnum = rdata.dat.shape[1]                                                 # number of traces in rgram 
    rdata.set_twtt(arr = f["Time"][:].flatten())                                    # set two way travel time
    rdata.dt = np.diff(rdata.get_twtt())[0]                                         # sampling interval, sec
    rdata.prf = f["param_records"]["radar"]["prf"][0][0]                            # pulse repitition frequency
    rdata.nchan = 1

    # store truncated samples
    try:
        rdata.truncs = f["Truncate_Bins"][:].flatten()[0].astype(int)
    except:
        rdata.truncs = 0

    # parse nav
    rdata.geocrs = navcrs
    rdata.xyzcrs = navparse.xyzsys[body]
    rdata.navdf = navparse.getnav_cresis_mat(fpath, navcrs, body)

    # pull surface two-way travel time and initilize horizon
    rdata.pick.srf = "srf"
    # get two-way travel times to lidar surface from CReSIS snow radar data
    twtt_surf = f["Surface"][:].flatten()
    rdata.pick.horizons["srf"] = utils.twtt2sample(twtt_surf, rdata.dt)
    # account for truncated samples above row zero in radargram
    rdata.pick.horizons["srf"] -= rdata.truncs
    rdata.set_srfElev() 

    # store pulse rep
    rdata.info["PRF [kHz]"] = rdata.prf * 1e-3

    f.close()                                                   # close the file

    rdata.check_attrs()

    return rdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
RAGU radar data processing class and tools
"""
### imports ###
from ragu.tools import utils
from ragu.nav import navparse
import pyproj
import numpy as np
import pandas as pd
import numpy.matlib as matlib
import matplotlib.pyplot as plt
import scipy.interpolate as interp
import scipy.signal as signal

class proc(object):
    def __init__(self):
        #: np.ndarray(snum x tnum), previously processed radar data (amp)
        self.prev_amp = None
        #: np.ndarray(snum x tnum), previously processed radar data (dB)
        self.prev_dB = None
        #: np.ndarray(snum x tnum), current processed radar data (amp)
        self.curr_amp = None
        #: np.ndarray(snum x tnum), current processed radar data (dB)
        self.curr_dB = None

    def set_prev_amp(self, amp):
        self.prev_amp = amp

    def get_prev_amp(self):
        return self.prev_amp

    def set_prev_dB(self, dB):
        self.prev_dB = dB

    def get_prev_dB(self):
        return self.prev_dB

    def set_curr_amp(self, amp):
        self.curr_amp = amp

    def get_curr_amp(self):
        return self.curr_amp

    def set_curr_dB(self, dB):
        self.curr_dB = dB

    def get_curr_dB(self):
        return self.curr_dB


def set_tzero(self):
    # get mean trace and find max sample and update sampzero flag
    if self.info["Signal Type"] == "Chirp":
        meanTrace = np.nanmean(np.abs(self.dat), axis=1)
        self.flags.sampzero = np.nanargmax(meanTrace)
    elif self.info["Signal Type"] == "Impulse":
        self.flags.sampzero = np.nanmean(utils.get_srf(np.abs(self.dat), "Impulse")).astype(int)

    if self.flags.sampzero > 0:
        self.tzero_shift()
        # log
        out = '# Time zero shifted to:\n# sample:\t {}\n# time:\t\t {} nanoseconds'\
        .format(self.flags.sampzero,(self.flags.sampzero * self.dt * 1e9))
        self.log("rdata.set_tzero()" + "\n" + out)
        print(out)

    return


def tzero_shift(self):
    # shift 2d proc data array so first row is time zero sample - use nan to fill bottom samples
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())
    out = np.zeros_like(amp)
    out[:-self.flags.sampzero,:] = amp[self.flags.sampzero:,:]
    out[-self.flags.sampzero:,:] = np.nan
    self.set_proc(out)

    return


def reverse(self):
    # reverse left-right order of radargram
    self.set_dat(self.get_dat()[:,::-1])
    self.set_proc(self.get_dat())

    # need to flip all relevant arrays - nav, picks, clutter
    if self.flags.sim:
        self.set_sim(utils.powdB2amp(self.sim)[:,::-1])

    # flip navdf and recalculate distance array
    self.navdf = self.navdf.iloc[::-1].reset_index(drop=True)
    self.navdf.dist = navparse.euclid_dist(self.navdf.x.to_numpy(), self.navdf.y.to_numpy(), self.navdf.z.to_numpy())

    # reverse picks
    for h in self.pick.horizons.keys():
        self.pick.horizons[h] = np.flip(self.pick.horizons[h])

    # log
    self.log("rdata.rgram_reverse()")
    print("# radargram reversed, to undo simply repeat reverse operation")

    return

def flatten(self):
    # flatten radargram by rolling each trace so that the surface is at sample zero
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())
    # initialize output array
    out = np.zeros_like(amp.T)
    # get surf samples in integer form for rolling
    self.flags.sampzero = self.pick.horizons[self.pick.get_srf()].astype(int)
    # loop through all traces and roll according to surface sample
    for i, col in enumerate(amp.T):
        out[i] = np.roll(col, shift = -self.flags.sampzero[i])
        out[i][-self.flags.sampzero[i]:] = np.nan             # set prior air samples to nan?

    self.set_proc(out.T)

    # log
    self.log("rdata.flatten()")
    print("# data array flattened ")

    return


def vertical_roll(self, samples=0):
    # roll 2d proc data array vertically to fix mismatch
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())
    out = np.roll(amp, shift=samples, axis=0)
    self.set_proc(out)

    # log
    self.log("rdata.vertical_roll(samples={})".format(samples))
    print("# data array rolled by {} samples".format(samples))

    return


def removeSlidingMeanFFT(self, window):
    # background noise removal using sliding mean in frequency space
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())

    # window in interval to avg over for mean removal
    mean = np.zeros(amp.shape)
    a = np.zeros(amp.shape[1])
    a[0 : window // 2] = 1
    a[amp.shape[1] - window // 2 : amp.shape[1]] = 1
    A = np.fft.fft(a)

    # Main circular convolution
    for i in range(amp.shape[0]):
        T = np.fft.fft(amp[i, :])
        mean[i, :] = np.real(np.fft.ifft(np.multiply(T, A)) / window)

    # Handle edges
    mt = np.zeros(amp.shape[0])
    for i in range(0, window):
        mt = np.add(mt, np.divide(amp[:, i], window))

    for i in range(0, window // 2):
        mean[:, i] = mt

    mt = np.zeros(amp.shape[0])
    for i in range(amp.shape[1] - window, amp.shape[1]):
        mt = np.add(mt, np.divide(amp[:, i], window))

    for i in range(amp.shape[1] - window // 2, amp.shape[1]):
        mean[:, i] = mt

    out = np.subtract(amp, mean)
    self.set_proc(out)
    # log
    self.log("rdata.removeSlidingMeanFFT(window={})".format(window))
    print("# Background removal completed wtih a window size of {} traces".format(window))

    return 


def butter(btype="lowpass", lowcut=None, highcut=None, fs=None, order=5):
    nyq = 0.5 * fs
    cutoff = []
    if btype=="lowpass" and highcut > 0:
        cutoff.append(highcut / nyq)
    elif btype=="highpass" and lowcut > 0:
        cutoff.append(lowcut / nyq)
    elif btype=="bandpass" and lowcut > 0 and highcut > 0:
        cutoff.append(lowcut / nyq)
        cutoff.append(highcut / nyq)
    else:
        raise ValueError("Critical frequency error: Lowcut={}, Highcut={}".format(lowcut, highcut))
        return
    b, a = signal.butter(order, cutoff, btype=btype)

    return b, a


def hilbertxform(self):
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())
    analytic_signal = signal.hilbert(amp, axis=0)
    amplitude_envelope = np.abs(analytic_signal)
    self.set_proc(amplitude_envelope)
    # log
    self.log("rdata.hilbertxform()")
    print("# hilbert transform applied applied")

    return 


def filter(self, btype="lowpass", lowcut=None, highcut=None, order=5, direction=0):
    # apply low pass filter to data array
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    # use abs value of amp
    amp = np.abs(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())
    if direction == 0:
        fs=1/self.dt
    elif direction == 1:
        fs=self.prf
    b, a = butter(btype=btype, lowcut=lowcut, highcut=highcut, fs=fs, order=order)
    # avoid scipy filter nan issues by windowing out any time zero shift
    out = np.zeros_like(amp)
    # get indices of any nans and temporarily replace
    idx = np.where(np.isnan(amp))
    amp[idx] = -9999
    out = signal.filtfilt(b, a, amp, axis=direction)
    out[idx] = np.nan
    # out[:-self.flags.sampzero,:] = signal.filtfilt(b, a, np.abs(amp[:-self.flags.sampzero:,:]), axis=direction)
    # out[-self.flags.sampzero:,:] = np.nan
    # use amplitude of lp filtered data to reset as pc array
    self.set_proc(out)
    # log
    self.log("rdata.filter(btype='{}', lowcut={}, highcut={}, order={}, direction={})".format(btype, lowcut, highcut, order, direction))
    print("# filter applied: btype='{}', lowcut={}, highcut={}, order={}, direction={}".format(btype, lowcut, highcut, order, direction))

    return


def restack(self, intrvl=None,thold=None):
    # get coordinate transformation
    xform = pyproj.Transformer.from_crs(self.geocrs, self.xyzcrs)

    # restack radar data to specified along-track distance
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    navdf = self.navdf.copy()
    # first account for any static traces where there may be gps drift
    if thold > 0:
        consecutive_dist = np.zeros_like(navdf.x)
        consecutive_dist[1:] = np.sqrt(np.diff(navdf.x.to_numpy()) ** 2.0 + np.diff(navdf.y.to_numpy()) ** 2.0 + np.diff(navdf.z.to_numpy()) ** 2.0)
        drift_mask = consecutive_dist >= thold

        # drop data from these traces
        namp = amp[:,drift_mask]
        navdf = navdf[drift_mask]
        navdf = navdf.reset_index(drop=True)
        # recalculcate distance after accounting for gps drift
        navdf["dist"] = navparse.euclid_dist(navdf["x"], navdf["y"], navdf["z"])
        self.snum, self.tnum = namp.shape
        amp = namp
        # get surface elev
        navdf['srfelev'] = self.srfElev[drift_mask]

    totdist = navdf.dist.iloc[-1]  # Total distance
    ntrace = int(totdist//intrvl)

    rstack = np.zeros((self.snum, ntrace))
    lat = np.zeros(ntrace)
    lon = np.zeros(ntrace)
    hgt = np.zeros(ntrace)
    srf = np.repeat(np.nan,ntrace)
    asep = np.repeat(np.nan,ntrace)
    twtt_wind = np.repeat(np.nan,ntrace)

    if "asep" not in navdf.keys():
        navdf["asep"] = self.asep
        
    for i in range(ntrace):
        stack_slice = np.logical_and(navdf.dist > i*intrvl, navdf.dist < (i+1)*intrvl)
        nstack = np.sum(stack_slice)
        if(nstack == 0):
            if i==0:
                rstack[:, i] = amp[:, i]
                lat[i] = navdf["lat"][i]
                lon[i] = navdf["lon"][i]
                hgt[i] = navdf["elev"][i]
                srf[i] = navdf["srfelev"][i]
                twtt_wind[i] = navdf["twtt_wind"][i]
                asep[i] = navdf["asep"][i]
            else:
                rstack[:, i] = rstack[:, i-1]
                lat[i] = lat[i-1]
                lon[i] = lon[i-1]
                hgt[i] = hgt[i-1]
                srf[i] = srf[i-1]
                twtt_wind[i] = twtt_wind[i-1]
                asep[i] = asep[i-1]
            continue

        rstack[:, i] = np.sum(amp[:, stack_slice], axis=1)/nstack
        lat[i] = np.mean(navdf["lat"][stack_slice])
        lon[i] = np.mean(navdf["lon"][stack_slice])
        hgt[i] = np.mean(navdf["elev"][stack_slice])
        srf[i] = np.mean(navdf["srfelev"][stack_slice])
        twtt_wind[i] = np.mean(navdf["twtt_wind"][stack_slice])
        asep[i] = np.mean(navdf["asep"][stack_slice])

    # store updated nav data
    self.navdf = pd.DataFrame()
    self.navdf["lon"] = lon
    self.navdf["lat"] = lat
    self.navdf["elev"] = hgt
    self.navdf["twtt_wind"] = twtt_wind
    self.navdf["asep"] = asep
    self.asep = asep

    self.set_srfElev(dat = srf)

    self.navdf["x"], self.navdf["y"], self.navdf["z"] = xform.transform(self.navdf["lon"], self.navdf["lat"], self.navdf["elev"])
    self.navdf["dist"] = navparse.euclid_dist(self.navdf["x"], self.navdf["y"], self.navdf["z"])

    self.snum, self.tnum = rstack.shape
    self.set_proc(rstack)
    # log
    self.log("rdata.restack(intrvl={},thold={})".format(intrvl,thold))
    print("# data restacked at an interval of {} m, with a minimum distance threshold of {} m".format(intrvl,thold))
    
    return

def tpowGain(self, power):
    # t-power gain to each trace with the given exponent.
    amp = self.proc.get_curr_amp()
    self.proc.set_prev_amp(amp)
    self.proc.set_prev_dB(self.proc.get_curr_dB())
    twtt = np.arange(self.snum)*self.dt
    factor = np.reshape(twtt**(float(power)),(len(twtt),1))
    factmat = np.matlib.repmat(factor,1,self.tnum)
    out = np.multiply(amp,factmat)
    self.set_proc(out)
    # log
    self.log("rdata.tpowGain(power={})".format(power))
    print("# time^{} gain applied".format(power))

    return


def undo(self):
    # undo last processing step
    if len(self.hist) > 2:
        tmp = self.proc.get_curr_amp()
        self.set_proc(self.proc.get_prev_amp())
        self.proc.set_prev_amp(tmp)
        # clear last log entry
        self.last = self.hist[-1]
        del self.hist[-1]
        print("# last processing step removed")

    return


def redo(self):
    tmp = self.proc.get_curr_amp()
    self.set_proc(self.proc.get_prev_amp())
    self.proc.set_prev_amp(tmp)
    self.log(self.last)

    return


def reset(self):
    # reset processed data to original
    if self.dtype == "oibak":
        self.set_proc(np.abs(self.dat))

    else:
        self.set_proc(self.dat)

    # clear log of all processing
    del self.hist[2:]
    
    return


### proc functions still in dev ###

def dewow(data,window):
    """
    Subtracts from each sample along each trace an 
    along-time moving average.

    Can be used as a low-cut filter.

    INPUT:
    data       data matrix whose columns contain the traces 
    window     length of moving average window 
            [in "number of samples"]

    OUTPUT:
    newdata    data matrix after dewow
    """
    totsamps = data.shape[0]
    # If the window is larger or equal to the number of samples,
    # then we can do a much faster dewow
    if (window >= totsamps):
        newdata = data-np.matrix.mean(data,0)            
    else:
        newdata = np.asmatrix(np.zeros(data.shape))
        halfwid = int(np.ceil(window/2.0))
        
        # For the first few samples, it will always be the same
        avgsmp=np.matrix.mean(data[0:halfwid+1,:],0)
        newdata[0:halfwid+1,:] = data[0:halfwid+1,:]-avgsmp

        # for each sample in the middle
        for smp in range(halfwid,totsamps-halfwid+1):
            winstart = int(smp - halfwid)
            winend = int(smp + halfwid)
            avgsmp = np.matrix.mean(data[winstart:winend+1,:],0)
            newdata[smp,:] = data[smp,:]-avgsmp

        # For the last few samples, it will always be the same
        avgsmp = np.matrix.mean(data[totsamps-halfwid:totsamps+1,:],0)
        newdata[totsamps-halfwid:totsamps+1,:] = data[totsamps-halfwid:totsamps+1,:]-avgsmp
        
    return newdata


def agcGain(data, window=50, scaling_factor=50):
    """Try to do some automatic gain control

    Parameters
    ----------
    window: int, optional
        The size of window we use in number of samples (default 50)
    scaling_factor: int, optional
        The scaling factor. This gets divided by the max amplitude when we rescale the input.
        Default 50.
    """
    num_sample = data.shape[1]
    maxamp = np.zeros((num_sample,))
    for i in range(num_sample):
        maxamp[i] = np.max(np.abs(data[max(0, i - window // 2):
                                            min(i + window // 2, num_sample), :]))
    maxamp[maxamp == 0] = 1.0e-6
    newdata = data * (scaling_factor / np.atleast_2d(maxamp).transpose()).astype(data.dtype)

    return newdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
pick export functions for RAGU
"""
### imports ###
from ragu.tools import utils
from ragu.raguError import raguError
from ragu.tools.constants import *
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
import tkinter as tk
import os, sys, h5py, fnmatch
import matplotlib.pyplot as plt

# pick_math is a function to perform all the necessary mathematics on a set of picks and save data as a pandas dataframe
# if overlapping pick segments exist, save as separate layers
def pick_math(rdata, i_eps_r=3.15, amp_out=True, horizon=None, srf=None):
    trace = np.arange(rdata.tnum)                       # array to hold trace number
    sample = rdata.pick.horizons.copy()
    # get list of horizon names
    horizons = list(rdata.pick.horizons.keys())
    damp = None

    # prep data amplitude array
    if (amp_out) and (rdata.dtype != "marsis"):
        # if raw data is complex, take absolute value to get amplitude
        if np.iscomplex(rdata.dat).all():
            damp = np.abs(rdata.dat)
        elif np.iscomplexobj(rdata.dat):
            damp = np.real(rdata.dat)
        else:
            damp = rdata.dat

    # initilize output dataframe
    out = pd.DataFrame({"trace": trace, 
                    "lon": rdata.navdf["lon"], 
                    "lat": rdata.navdf["lat"], 
                    "elev": rdata.navdf["elev"]})

    ### export single horizon ### - trace, lon, lat, elev, sample, twtt, amp
    if horizon and horizon in horizons:
        # apply sample time zero shift back in
        samp_arr = sample[horizon] + rdata.flags.sampzero
        out["sample"] = samp_arr
        # get corresponding twtt - account for data truncation
        twtt_arr = np.repeat(np.nan, rdata.tnum)
        idx = ~np.isnan(samp_arr)
        twtt_arr[idx] = rdata.get_twtt()[samp_arr[idx].astype(int) +  rdata.truncs]
        out["twtt"] = twtt_arr
        # add in twtt_wind to get absolute twtt
        out["twtt"] += rdata.navdf["twtt_wind"]

        if type(damp) is np.ndarray:
            amp = np.repeat(np.nan, rdata.tnum)
            idx = ~np.isnan(samp_arr)
            amp[idx] = damp[samp_arr[idx].astype(int), idx]
            out["amp"] = amp
        return out

    ### export merged horizons ### - reference surface elevation if present
    if horizon is None:
        # iterate through interpretation horizons
        for i, (horizon, array) in enumerate(sample.items()):
            samp_arr = array + rdata.flags.sampzero
            out[horizon + "_sample"] = samp_arr
            # get corresponding twtt - account for data truncation
            twtt_arr = np.repeat(np.nan, rdata.tnum)
            idx = ~np.isnan(samp_arr)
            twtt_arr[idx] = rdata.get_twtt()[(samp_arr[idx].astype(int)) + rdata.truncs]
            out[horizon + "_twtt"] = twtt_arr
            # add in twtt_wind to get absolute twtt
            out[horizon + "_twtt"] += rdata.navdf["twtt_wind"]


            if horizon == srf:
                out[horizon + "_elev"] = rdata.get_srfElev()

            if type(damp) is np.ndarray:
                # export horizon amplitude values
                amp = np.repeat(np.nan, rdata.tnum)
                idx = ~np.isnan(samp_arr)
                # add any applied shift to index to pull proper sample amplitude from data array
                amp[idx] = damp[samp_arr[idx].astype(int), idx]
                out[horizon + "_amp"] = amp

            if i > 0:
                # get thickness between current layer and preceding layer
                # first confirm users preferred dielectric permittivity for layer
                eps_r = None
                while (eps_r is None) or (eps_r < 1):
                    eps_r = tk.simpledialog.askfloat("Dielectric Permittivity","Select a relative dielectric permittivity\nfor the unit between horizon <{}> and horizon <{}>".format(horizons[i - 1], horizons[i]), initialvalue=i_eps_r) 
                    if eps_r is None:
                        return None
                    elif eps_r < 1:
                        print("raguWarning: A relative dielectric permittivity >=1 must be specified in order to export picks")

                h = utils.twtt2depth(out[horizon + "_twtt"] - out[horizons[i - 1] + "_twtt"], rdata.asep, eps_r)
                # calculate layer bed elevation as elevation of preceding layer minus layer thickness - this only works if a surface with reference elevation is defined
                if srf:
                    out[horizon + "_elev"] = out[horizons[i - 1] + "_elev"] - h
                out[horizons[i - 1] + "_" + horizon + "_thick"] = h

        return out


# csv is a function to export the output pick dataframe as a csv
def csv(fpath, df):
    # fpath is the path for where the exported csv pick file should be saved [str]
    # df pick output dataframe
    if isinstance(df,pd.DataFrame):

        df.to_csv(fpath, index=False)

        print("csv picks exported successfully:\t" + fpath)


# gpkg is a funciton for saving picks to a geopackage/shapefile
def gpkg(fpath, df, crs):
    # fpath is the path for where the exported csv pick file should be saved [str]
    # df pick output dataframe
    # crs is the coordinate reference system for the shapefile output
    if isinstance(df,pd.DataFrame):
        df_copy = df.copy()
        if df_copy["lon"].isnull().all() or df_copy["lat"].isnull().all():
            print("no geopackage was exported due to missing gps data")
            return
        # convert lon, lat to shapely points
        geometry = [Point(xy) for xy in zip(df_copy["lon"], df_copy["lat"])]
        df_copy.drop(["lon", "lat"], axis=1)

        # create geopandas df and export
        gdf = gpd.GeoDataFrame(df, crs=crs, geometry=geometry)
        gdf.to_file(fpath, driver="GPKG")

        print("geopackage exported successfully:\t" + fpath)


# h5 is a function for saving twtt_bed pick to h5 data file
def h5(fpath, df=None, dtype=None, srf=None):
    # fpath is the data file path [str]
    # df pick output dataframe
    if (dtype=="oibak"):
        f = h5py.File(fpath, "a")
        flag = False
        # update twtt_surf
        if srf:
            dat = df[srf + "_twtt"].to_numpy()
            # if twtt_surf exists in file, see it current picks have been modified
            if "twtt_surf" in f["drv"]["pick"].keys():
                twtt_srf_dfile = f["drv"]["pick"]["twtt_surf"][:]
                twtt_srf_dfile[twtt_srf_dfile == -1] = np.nan
                twtt_srf_dfile[twtt_srf_dfile == -9] = np.nan
                if not utils.nan_array_equal(twtt_srf_dfile, dat) or (np.isnan(twtt_srf_dfile).all()):
                    if (tk.messagebox.askyesno("twtt_surf","Export twtt_surf pick layer to data file?")):
                        del f["drv"]["pick"]["twtt_surf"]
                        flag = True
            elif (tk.messagebox.askyesno("twtt_surf","Export twtt_surf pick layer to data file?")):
                flag = True
            if flag:
                dat[np.isnan(dat)] = -9
                twtt_surf = f["drv"]["pick"].require_dataset("twtt_surf", data=dat, shape=dat.shape, dtype=np.float32)
                # twtt_surf.attrs.create("Unit", np.string_("Seconds"))
                # twtt_surf.attrs.create("Source", np.string_("Manual pick layer"))
                print("twtt_surf exported successfully:\t" + fpath + "/drv/pick/twtt_surf")

        # update twtt_bed
        if "bed_twtt" in df.keys():
            dat = df["bed_twtt"].to_numpy()
            dat[np.isnan(dat)] = -9
            if "twtt_bed" in f["drv"]["pick"].keys():
                if (f["drv"]["pick"]["twtt_bed"][:] == dat).all():
                    return
                elif tk.messagebox.askyesno("twtt_bed","Export twtt_bed pick layer to data file?"):
                    del f["drv"]["pick"]["twtt_bed"]
            elif not tk.messagebox.askyesno("twtt_bed","Export twtt_bed pick layer to data file?"):
                return
            twtt_bed = f["drv"]["pick"].require_dataset("twtt_bed", data=dat, shape=dat.shape, dtype=np.float32)
            # twtt_bed.attrs.create("Unit", np.string_("Seconds"))
            # twtt_bed.attrs.create("Source", np.string_("Manual pick layer"))
            print("twtt_bed exported successfully:\t\t" + fpath + "/drv/pick/twtt_bed")
        f.close()

    else:
        return


# fig is a function for exporting the pick image
def fig(fpath, fig):
    fig.savefig(fpath, dpi=500, bbox_inches='tight', pad_inches=0.05, transparent=True)# facecolor = "#d9d9d9")
    print("figure exported successfully:\t" + fpath)


# dat is a method to export the radar data - for now just as a csv file array
def dat(fpath, dat):
    np.savetxt(fpath, dat, fmt="%s", delimiter=",")
    print("data exported successfully:\t" + fpath)


# log is a method to export the processing log as a python script
def log(fpath, log):
    with open(fpath,"w") as ofile:
        ofile.write("### RAGU processing log ###\nfrom ragu import ingest\n\n")
        for _i in log:
            ofile.write(_i + "\n")
        ofile.close()


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_marsis is a module developed to ingest JPL MARSIS radar sounding data. 
data format is binary 32-bit floating point pulse compressed data
"""
### imports ###
from ragu.radar import garlic
from ragu.nav import navparse
from ragu.tools import utils
from PIL import Image
import numpy as np
import os, sys, glob
import matplotlib.pyplot as plt

# method to read JPL multilook MARSIS data
def read(fpath, simpath, navcrs, body):
    fn = fpath.split("/")[-1]
    root = fpath.rstrip(fn)
    orbit = fn.split('_')
    orbit = orbit[0] + '_' + orbit[1]
    rdata = garlic(fpath)
    rdata.fn = fn[:-4]
    rdata.dtype = "marsis"

    # convert binary RGRAM to numpy array
    # # reshape array
    dtype = np.dtype("float32")     
    with open(fpath, "rb") as f:
        dat = np.fromfile(f, dtype)     
    l = len(dat)

    rdata.snum = 2048
    # get number of traces, dividing file length by number of samples per trace, by 8 data arrays
    rdata.tnum = int(l/rdata.snum/8)
    # dt per pixel from reprocessed oversampled data - data is oversampled by factor of 2
    rdata.fs = 2*(1.4e6)
    rdata.dt = 1/rdata.fs
    rdata.prf = 127
    rdata.nchan = 2

    # reshape into 8 stacked rgrams
    dat = dat.reshape((rdata.snum*8,rdata.tnum),order="F")

    # reprocessed MARSIS data should be bottom two rgrams
    dat = dat[-(rdata.snum*rdata.nchan):,:]

    # reshape into stacked 3D array for two channels
    tmp = np.zeros((rdata.snum,rdata.tnum,rdata.nchan))
    tmp[:,:,0] = dat[:rdata.snum,:]
    tmp[:,:,1] = dat[-rdata.snum:,:]
    # apparently data arrays are already power values, so revert to amplitude (abs(amplitude))
    rdata.set_dat(np.sqrt(tmp))
    rdata.set_proc(rdata.get_dat())

    # convert png clutter sim product to numpy array
    if simpath:
        simpath = simpath + "/" + orbit + "_clutter.img"
    else:
        simpath = root + "/" + orbit + "_clutter.img"

    if os.path.isfile(simpath):
        with open(simpath,"rb") as f:
            dat = np.fromfile(f, np.uint8)
            sim = dat.reshape(rdata.snum, rdata.tnum)
            rdata.set_sim(sim)
    else:
        print("Clutter simulation not found:\t{}\nSpecify alternate path in configuration file.".format(simpath))

    rdata.set_twtt()

    # assign signal info
    rdata.info["Signal Type"] = "Chirp"
    rdata.info["Sampling Frequency [MHz]"] = rdata.fs*1e-6
    rdata.info["PRF [Hz]"] = rdata.prf

    # open geom nav file for rgram
    geom_path = root + orbit + "_geom.tab"
 
    # parse nav
    rdata.geocrs = navcrs
    rdata.xyzcrs = navparse.xyzsys[body]
    rdata.navdf = navparse.getnav_marsis(geom_path, navcrs, body)

    rdata.set_srfElev(dat = np.repeat(np.nan, rdata.tnum))

    rdata.check_attrs()

    return rdata


# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
"""
ingest_sharad is a module developed to ingest NASA MRO-SHARAD FPB radar sounding data. 
data format is binary 32-bit floating point pulse compressed amplitude data acquired from the PDS
"""
### imports ###
from ragu.radar import garlic
from ragu.nav import navparse
from ragu.tools import utils
import numpy as np
import os, sys

# method to read PDS SHARAD USRDR data
def read(fpath, simpath, navcrs, body):
    rdata = garlic(fpath)
    if fpath.endswith("sim.img") or fpath.endswith("geom_combined.img"):
        return rdata
    rdata.fn = fpath.split("/")[-1][:-10]
    rdata.dtype = "sharad"
    root = os.path.dirname(fpath)

    # rdata.fn = fn.rstrip("_rgram.img")
    # convert binary .img PDS RGRAM to numpy array
    # reshape array with 3600 lines
    dtype = np.dtype("float32")     
    with open(fpath, "rb") as f:
        tmp = np.fromfile(f, dtype)     
    l = len(tmp)

    rdata.snum = 3600
    rdata.tnum = int(len(tmp)/rdata.snum)
    rdata.dt = .0375e-6
    rdata.prf = 700.28
    rdata.nchan = 1
    rdata.set_dat(tmp.reshape(rdata.snum,rdata.tnum))
    rdata.set_proc(rdata.get_dat())

    # convert binary .img clutter sim product to numpy array
    # account for PDS v4 sim format and legacy geom_combined format
    if simpath:
        simpath = simpath + "/" + rdata.fn + "_sim.img"
        if not os.path.isfile(simpath):
            simpath = simpath + "/" + rdata.fn + "_geom_combined.img"

    else:
        simpath = root + "/" + rdata.fn + "_sim.img"
        if not os.path.isfile(simpath):
            simpath = root + "/" + rdata.fn + "_geom_combined.img"

    if os.path.isfile(simpath):
        with open(simpath, "rb") as f:
            sim = np.fromfile(f, dtype)
        # reshape - will be different depending on sim version
        if simpath.endswith('sim.img'):
            # just take combined sim if PDS v4 sim
            sim = sim[2 * len(sim) // 3 :].reshape(rdata.snum,rdata.tnum)
        else:
            sim = sim.reshape(rdata.snum, rdata.tnum)
    
        rdata.set_sim(sim)
    else:
        print("Clutter simulation not found in:\t{}\nSpecify alternate path in configuration file.".format(simpath.rstrip(simpath.split('/')[-1])))

    rdata.set_twtt()

    # assign signal info
    rdata.info["signal type"] = "chirp"
    rdata.info["cf [MHz]"] = 20
    rdata.info["badwidth [%]"] = 50
    rdata.info["pulse length [\u03BCs]"] = 85
    rdata.info["prf [Hz]"] = rdata.prf

    # open geom nav file for rgram
    geom_path = root +  "/" + rdata.fn + "_geom.tab"

    # parse nav
    rdata.geocrs = navcrs
    rdata.xyzcrs = navparse.xyzsys[body]
    rdata.navdf = navparse.getnav_sharad(geom_path, navcrs, body)

    rdata.set_srfElev(dat = np.repeat(np.nan, rdata.tnum))

    rdata.check_attrs()

    return rdata